"""Contains all classes related to the Character Bazaar sections in Tibia.com"""
import datetime
import logging
import re
import urllib.parse
from typing import Dict, List, Optional

import bs4

from tibiapy import InvalidContent, Sex, Vocation, abc
from tibiapy.abc import BaseCharacter
from tibiapy.enums import (AuctionOrder, AuctionOrderBy, AuctionSearchType, AuctionStatus, BattlEyeTypeFilter,
                           BazaarType, BidType, PvpTypeFilter, SkillFilter, VocationAuctionFilter)
from tibiapy.utils import (convert_line_breaks, get_tibia_url, parse_form_data, parse_integer, parse_pagination,
                           parse_tibia_datetime, parse_tibiacom_content, try_enum)

__all__ = (
    "AchievementEntry",
    "Auction",
    "AuctionFilters",
    "CharacterBazaar",
    "CharmEntry",
    "BestiaryEntry",
    "BlessingEntry",
    "DisplayItem",
    "DisplayMount",
    "DisplayOutfit",
    "DisplayFamiliar",
    "ItemSummary",
    "AuctionEntry",
    "Outfits",
    "OutfitImage",
    "Mounts",
    "Familiars",
    "SalesArgument",
    "SkillEntry",
)

results_pattern = re.compile(r'Results: (\d+)')
char_info_regex = re.compile(r'Level: (\d+) \| Vocation: ([\w\s]+)\| (\w+) \| World: (\w+)')
id_addon_regex = re.compile(r'(\d+)_(\d)\.gif')
id_regex = re.compile(r'(\d+).(?:gif|png)')
description_regex = re.compile(r'"(?:an?\s)?([^"]+)"')
amount_regex = re.compile(r'([\d,]+)x')

log = logging.getLogger("tibiapy")


class AchievementEntry(abc.Serializable):
    """An unlocked achievement by the character.

    Attributes
    ----------
    name: :class:`str`
        The name of the achievement.
    secret: :class:`bool`
        Whether the achievement is secret or not.
    """

    def __init__(self, name, secret=False):
        self.name: str = name
        self.secret: int = secret

    __slots__ = (
        "name",
        "secret",
    )

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r} secret={self.secret}>"


class AuctionFilters(abc.Serializable):
    """The auction filters available in the auctions section.

    All attributes are optional.

    Attributes
    ----------
    world: :class:`str`
        The character's world to show characters for.
    pvp_type: :class:`PvpTypeFilter`
        The PvP type of the character's worlds to show.
    battleye: :class:`BattlEyeTypeFilter`
        The type of BattlEye protection of the character's worlds to show.
    vocation: :class:`VocationAuctionFilter`
        The character vocation to show results for.
    min_level: :class:`int`
        The minimum level to display.
    max_level: :class:`int`
        The maximum level to display.
    skill: :class:`SkillFilter`
        The skill to filter by its level range.
    min_skill_level: :class:`int`
        The minimum skill level of the selected :attr:`skill` to display.
    max_skill_level: :class:`int`
        The maximum skill level of the selected :attr:`skill` to display.
    search_string: :class:`str`
        The search term to filter out auctions.
    search_type: :class:`AuctionSearchType`
        The type of search to use. Defines the behaviour of :py:attr:`search_string`.
    available_worlds: :class:`list` of :class:`str`
        The list of available worlds to select to filter.
    """

    __slots__ = (
        "world",
        "pvp_type",
        "battleye",
        "vocation",
        "min_level",
        "max_level",
        "skill",
        "min_skill_level",
        "max_skill_level",
        "order_by",
        "order",
        "search_string",
        "search_type",
        "available_worlds",
    )

    def __init__(self, **kwargs):
        self.world: Optional[str] = kwargs.get("world")
        self.pvp_type: Optional[PvpTypeFilter] = kwargs.get("pvp_type")
        self.battleye: Optional[BattlEyeTypeFilter] = kwargs.get("battleye")
        self.vocation: Optional[VocationAuctionFilter] = kwargs.get("vocation")
        self.min_level: Optional[int] = kwargs.get("min_level")
        self.max_level: Optional[int] = kwargs.get("max_level")
        self.skill: Optional[SkillFilter] = kwargs.get("skill")
        self.min_skill_level: Optional[int] = kwargs.get("min_skill_level")
        self.max_skill_level: Optional[int] = kwargs.get("max_skill_level")
        self.order_by: Optional[AuctionOrderBy] = kwargs.get("order_by")
        self.order: Optional[AuctionOrder] = kwargs.get("order")
        self.search_string: Optional[str] = kwargs.get("search_string")
        self.search_type: Optional[AuctionSearchType] = kwargs.get("search_type")
        self.available_worlds: List[str] = kwargs.get("available_worlds", [])

    def __repr__(self):
        attributes = ""
        for attr in self.__slots__:
            v = getattr(self, attr)
            attributes += f" {attr}={v!r}"
        return f"<{self.__class__.__name__}{attributes}>"

    @property
    def query_params(self):
        """:class:`dict`: The query parameters representing this filter."""
        params = {
            "filter_profession": self.vocation.value if self.vocation else None,
            "filter_levelrangefrom": self.min_level,
            "filter_levelrangeto": self.max_level,
            "filter_world": self.world,
            "filter_worldpvptype": self.pvp_type.value if self.pvp_type else None,
            "filter_worldbattleyestate": self.battleye.value if self.battleye else None,
            "filter_skillid": self.skill.value if self.skill else None,
            "filter_skillrangefrom": self.min_skill_level,
            "filter_skillrangeto": self.max_skill_level,
            "order_column": self.order_by.value if self.order_by else None,
            "order_direction": self.order.value if self.order else None,
            "searchstring": self.search_string,
            "searchtype": self.search_type.value if self.search_type else None,
        }
        return {k: v for k, v in params.items() if v is not None}

    @classmethod
    def _parse_filter_table(cls, table):
        """Parse the filters table to extract its values.

        Parameters
        ----------
        table: :class:`bs4.Tag`
            The table containing the filters.

        Returns
        -------
        :class:`AuctionFilters`
            The currently applied filters.
        """
        filters = AuctionFilters()
        forms = table.find_all("form")
        data = parse_form_data(forms[0], include_options=True)

        filters.world = data["filter_world"]
        filters.available_worlds = [w for w in data.get("__options__", {}).get("filter_world", []) if "(" not in w]
        filters.pvp_type = try_enum(PvpTypeFilter, parse_integer(data.get("filter_worldpvptype"), None))
        filters.battleye = try_enum(BattlEyeTypeFilter, parse_integer(data.get("filter_worldbattleyestate"), None))
        filters.vocation = try_enum(VocationAuctionFilter, parse_integer(data.get("filter_profession"), None))
        filters.min_level = parse_integer(data.get("filter_levelrangefrom"), None)
        filters.max_level = parse_integer(data.get("filter_levelrangeto"), None)
        filters.skill = try_enum(SkillFilter, parse_integer(data.get("filter_skillid"), None))
        filters.min_skill_level = parse_integer(data.get("filter_skillrangefrom"), None)
        filters.max_skill_level = parse_integer(data.get("filter_skillrangeto"), None)
        filters.order_by = try_enum(AuctionOrderBy, parse_integer(data.get("order_column"), None))
        filters.order = try_enum(AuctionOrder, parse_integer(data.get("order_direction"), None))
        if len(forms) > 1:
            data_search = parse_form_data(forms[1], include_options=True)
            filters.search_string = data_search.get("searchstring")
            filters.search_type = try_enum(AuctionSearchType, parse_integer(data_search.get("searchtype"), None))
        return filters


class BestiaryEntry(abc.Serializable):
    """The bestiary progress for a specific creature.

    Attributes
    ----------
    name: :class:`str`
        The name of the creature.
    kills: :class:`int`
        The number of kills of this creature the player has done.
    step: :class:`int`
        The current step to unlock this creature the character is in, where 4 is fully unlocked.
    """

    def __init__(self, name, kills, step):
        self.name: str = name
        self.kills: int = kills
        self.step: int = step

    __slots__ = (
        "name",
        "kills",
        "step",
    )

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r} kills={self.kills} step={self.step}>"

    @property
    def completed(self):
        """:class:`bool`: Whether the entry is completed or not."""
        return self.step == 4


class BlessingEntry(abc.Serializable):
    """A character's blessings.

    Attributes
    ----------
    name: :class:`str`
        The name of the blessing.
    amount: :class:`int`
        The amount of blessing charges the character has.
    """

    __slots__ = (
        "name",
        "amount",
    )

    def __init__(self, name, amount=0):
        self.name: str = name
        self.amount: int = amount

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r} amount={self.amount}>"


class CharacterBazaar(abc.Serializable):
    """Represents the char bazaar.

    Attributes
    ----------
    page: :class:`int`
        The page being currently viewed.
    total_pages: :class:`int`
        The total number of pages available.
    results_count: :class:`int`
        The number of auctions listed.
    entries: :class:`list` of :class:`AuctionEntry`
        The auctions displayed.
    type: :class:`BazaarType`
        The type of auctions being displayed, either current or auction history.
    filters: :class:`AuctionFilters`
        The currently set filtering options.
    """

    __slots__ = (
        "type",
        "filters",
        "page",
        "total_pages",
        "results_count",
        "entries",
    )

    def __init__(self, **kwargs):
        self.type: BazaarType = kwargs.get("type")
        self.filters: Optional[AuctionFilters] = kwargs.get("filters")
        self.page: int = kwargs.get("page", 1)
        self.total_pages: int = kwargs.get("total_pages", 1)
        self.results_count: int = kwargs.get("results_count", 0)
        self.entries: List[AuctionEntry] = kwargs.get("entries", [])

    def __repr__(self):
        return f"<{self.__class__.__name__} page={self.page} total_pages={self.total_pages} " \
               f"results_count={self.results_count}>"

    # region Properties
    @property
    def url(self):
        """:class:`str`: Get the URL to the bazaar."""
        return self.get_auctions_history_url(self.page) if self.type == BazaarType.HISTORY else \
            self.get_current_auctions_url(self.page, self.filters)
    # endregion

    # region Public Methods
    @classmethod
    def get_current_auctions_url(cls, page=1, filters=None):
        """Get the URL to the list of current auctions in Tibia.com.

        Parameters
        ----------
        page: :class:`int`
            The page to show the URL for.
        filters: :class:`AuctionFilters`
            The filtering criteria to use.

        Returns
        -------
        :class:`str`
            The URL to the current auctions section in Tibia.com
        """
        filters = filters or AuctionFilters()
        return get_tibia_url("charactertrade", "currentcharactertrades", currentpage=page, **filters.query_params)

    @classmethod
    def get_auctions_history_url(cls, page=1, filters=None):
        """Get the URL to the auction history in Tibia.com.

        Parameters
        ----------
        page: :class:`int`
            The page to show the URL for.
        filters: :class:`AuctionFilters`
            The filtering criteria to use.

        Returns
        -------
        :class:`str`
            The URL to the auction history section in Tibia.com
        """
        filters = filters or AuctionFilters()
        return get_tibia_url("charactertrade", "pastcharactertrades", currentpage=page, **filters.query_params)

    @classmethod
    def from_content(cls, content):
        """Get the bazaar's information and list of auctions from Tibia.com.

        Parameters
        ----------
        content: :class:`str`
            The HTML content of the bazaar section at Tibia.com.

        Returns
        -------
        :class:`CharacterBazaar`
            The character bazaar with the entries found.
        """
        try:
            parsed_content = parse_tibiacom_content(content, builder='html5lib')
            content_table = parsed_content.find("div", attrs={"class": "BoxContent"})
            tables = content_table.find_all("div", attrs={"class": "TableContainer"})
            filter_table = None
            if len(tables) == 1:
                auctions_table = tables[0]
            else:
                filter_table, auctions_table, *_ = tables

            bazaar = cls()
            bazaar.type = BazaarType.CURRENT if filter_table else BazaarType.HISTORY

            if filter_table:
                bazaar.filters = AuctionFilters._parse_filter_table(filter_table)

            page_navigation_row = parsed_content.find("td", attrs={"class": "PageNavigation"})
            if page_navigation_row:
                bazaar.page, bazaar.total_pages, bazaar.results_count = parse_pagination(page_navigation_row)

            auction_rows = auctions_table.find_all("div", attrs={"class": "Auction"})
            for auction_row in auction_rows:
                auction = AuctionEntry._parse_auction(auction_row)

                bazaar.entries.append(auction)
            return bazaar
        except ValueError as e:
            raise InvalidContent("content does not belong to the bazaar at Tibia.com", original=e)
    # endregion


class CharmEntry(abc.Serializable):
    """An unlocked charm by the character.

    Attributes
    ----------
    name: :class:`str`
        The name of the charm.
    cost: :class:`int`
        The cost of the charm in charm points.
    """

    def __init__(self, name, cost=0):
        self.name: str = name
        self.cost: int = cost

    __slots__ = (
        "name",
        "cost",
    )

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r} cost={self.cost}>"


class DisplayImage(abc.Serializable):
    """An image displayed in the auction.

    Attributes
    ----------
    image_url: :class:`str`
        The URL to the image.
    name: :class:`str`
        The element's name.
    """

    def __init__(self, **kwargs):
        self.image_url: str = kwargs.get("image_url")
        self.name: str = kwargs.get("name")

    __slots__ = (
        "image_url",
        "name",
    )

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r} image_url={self.image_url!r}>"

    @classmethod
    def _parse_image_box(cls, item_box):
        description = item_box["title"]
        img_tag = item_box.find("img")
        if not img_tag:
            return None
        return cls(image_url=img_tag["src"], name=description)


class DisplayItem(abc.Serializable):
    """Represents an item displayed on an auction, or the character's items in the auction detail.

    Attributes
    ----------
    image_url: :class:`str`
        The URL to the item's image.
    name: :class:`str`
        The item's name.
    description: :class:`str`
        The item's description, if any.
    count: :class:`int`
        The item's count.
    item_id: :class:`int`
        The item's client id.
    """

    __slots__ = (
        "image_url",
        "name",
        "description",
        "count",
        "item_id",
    )

    def __init__(self, **kwargs):
        self.image_url: str = kwargs.get("image_url")
        self.name: str = kwargs.get("name")
        self.description: str = kwargs.get("description")
        self.count: int = kwargs.get("count", 1)
        self.item_id: int = kwargs.get("item_id", 0)

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r} count={self.count} item_id={self.item_id}>"

    @classmethod
    def _parse_image_box(cls, item_box):
        title_text = item_box["title"]
        img_tag = item_box.find("img")
        if not img_tag:
            return None
        m = amount_regex.match(title_text)
        amount = 1
        if m:
            amount = parse_integer(m.group(1))
            title_text = amount_regex.sub("", title_text, 1).strip()
        item_id = 0
        description = None
        name, *desc = title_text.split("\n")
        if desc:
            description = " ".join(desc)
        m = id_regex.search(img_tag["src"])
        if m:
            item_id = int(m.group(1))
        return DisplayItem(image_url=img_tag["src"], name=name, count=amount, item_id=item_id, description=description)


class DisplayMount(DisplayImage):
    """Represents a mount owned or unlocked by the character.

    Attributes
    ----------
    image_url: :class:`str`
        The URL to the image.
    name: :class:`str`
        The mount's name.
    mount_id: :class:`int`
        The internal ID of the mount.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.mount_id: int = kwargs.get("mount_id", 0)

    __slots__ = (
        "mount_id",
    )

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r} mount_id={self.mount_id} image_url={self.image_url!r}>"

    @classmethod
    def _parse_image_box(cls, item_box):
        mount = super()._parse_image_box(item_box)
        m = id_regex.search(mount.image_url)
        if m:
            mount.mount_id = int(m.group(1))
        return mount


class DisplayOutfit(DisplayImage):
    """Represents an outfit owned or unlocked by the character.

    Attributes
    ----------
    image_url: :class:`str`
        The URL to the image.
    name: :class:`str`
        The outfit's name.
    outfit_id: :class:`int`
        The internal ID of the outfit.
    addons: :class:`int`
        The unlocked or owned addons for this outfit.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.outfit_id: int = kwargs.get("outfit_id", 0)
        self.addons: int = kwargs.get("addons", 0)

    __slots__ = (
        "outfit_id",
        "addons",
    )

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r} outfit_id={self.outfit_id} addons={self.addons} " \
               f"image_url={self.image_url!r}>"

    @classmethod
    def _parse_image_box(cls, item_box):
        outfit = super()._parse_image_box(item_box)
        name = outfit.name.split("(")[0].strip()
        outfit.name = name
        m = id_addon_regex.search(outfit.image_url)
        if m:
            outfit.outfit_id = int(m.group(1))
            outfit.addons = int(m.group(2))
        return outfit


class DisplayFamiliar(DisplayImage):
    """Represents a familiar owned or unlocked by the character.

    Attributes
    ----------
    image_url: :class:`str`
        The URL to the image.
    name: :class:`str`
        The familiar's name.
    familiar_id: :class:`int`
        The internal ID of the familiar.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.familiar_id: int = kwargs.get("familiar_id", 0)

    __slots__ = (
        "familiar_id",
    )

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r} familiar_id={self.familiar_id} " \
               f"image_url={self.image_url!r}>"

    @classmethod
    def _parse_image_box(cls, item_box):
        familiar = super()._parse_image_box(item_box)
        name = familiar.name.split("(")[0].strip()
        familiar.name = name
        m = id_regex.search(familiar.image_url)
        if m:
            familiar.familiar_id = int(m.group(1))
        return familiar


class AuctionEntry(BaseCharacter, abc.Serializable):
    """Represents an auction in the list, containing the summary.

    Attributes
    ----------
    auction_id: :class:`int`
        The internal id of the auction.
    name: :class:`str`
        The name of the character.
    level: :class:`int`
        The level of the character.
    world: :class:`str`
        The world the character is in.
    vocation: :class:`Vocation`
        The vocation of the character.
    sex: :class:`Sex`
        The sex of the character.
    outfit: :class:`OutfitImage`
        The current outfit selected by the user.
    displayed_items: :class:`list` of :class:`DisplayItem`
        The items selected to be displayed.
    sales_arguments: :class:`list` of :class:`SalesArgument`
        The sale arguments selected for the auction.
    auction_start: :class:`datetime.datetime`
        The date when the auction started.
    auction_end: :class:`datetime.datetime`
        The date when the auction ends.
    bid: :class:`int`
        The current bid in Tibia Coins.
    bid_type: :class:`BidType`
        The type of the auction's bid.
    status: :class:`AuctionStatus`
        The current status of the auction.
    """

    __slots__ = (
        "auction_id",
        "name",
        "level",
        "world",
        "vocation",
        "sex",
        "outfit",
        "displayed_items",
        "sales_arguments",
        "auction_start",
        "auction_end",
        "bid",
        "bid_type",
        "status",
    )

    def __init__(self, **kwargs):
        self.auction_id: int = kwargs.get("auction_id", 0)
        self.name: str = kwargs.get("name")
        self.level: int = kwargs.get("level", 0)
        self.world: str = kwargs.get("world")
        self.vocation: Vocation = kwargs.get("vocation")
        self.sex: Sex = kwargs.get("sex")
        self.outfit: OutfitImage = kwargs.get("outfit")
        self.displayed_items: List[DisplayItem] = kwargs.get("displayed_items", [])
        self.sales_arguments: List[SalesArgument] = kwargs.get("sales_arguments", [])
        self.auction_start: datetime.datetime = kwargs.get("auction_start")
        self.auction_end: datetime.datetime = kwargs.get("auction_end")
        self.bid: int = kwargs.get("bid", 0)
        self.bid_type: BidType = kwargs.get("bid_type")
        self.status: AuctionStatus = kwargs.get("status")

    def __repr__(self):
        return f"<{self.__class__.__name__} auction_id={self.auction_id} name={self.name!r} world={self.world!r}>"

    # region Properties
    @property
    def character_url(self):
        """:class:`str`: The URL of the character's information page on Tibia.com."""
        return BaseCharacter.get_url(self.name)

    @property
    def url(self):
        """:class:`str`: The URL to this auction's detail page on Tibia.com."""
        return self.get_url(self.auction_id)
    # endregion

    # region Public Methods
    @classmethod
    def get_url(cls, auction_id):
        """Get the URL to the Tibia.com detail page of an auction with a given id.

        Parameters
        ----------
        auction_id: :class:`int`
            The ID of the auction.

        Returns
        -------
        :class:`str`
            The URL to the auction's detail page.
        """
        return get_tibia_url("charactertrade", "currentcharactertrades", page="details", auctionid=auction_id)
    # endregion

    # region Private Methods
    @classmethod
    def _parse_auction(cls, auction_row, auction_id=0):
        """Parse an auction's table, extracting its data.

        Parameters
        ----------
        auction_row: :class:`bs4.Tag`
            The row containing the auction's information.
        auction_id: :class:`int`
            The ID of the auction.

        Returns
        -------
        :class:`AuctionEntry`
            The auction contained in the table.
        """
        header_container = auction_row.find("div", attrs={"class": "AuctionHeader"})
        char_name_container = header_container.find("div", attrs={"class": "AuctionCharacterName"})
        char_link = char_name_container.find("a")
        if char_link:
            url = urllib.parse.urlparse(char_link["href"])
            query = urllib.parse.parse_qs(url.query)
            auction_id = int(query["auctionid"][0])
            name = char_link.text
        else:
            name = char_name_container.text

        auction = cls(name=name, auction_id=auction_id)
        char_name_container.replaceWith('')
        m = char_info_regex.search(header_container.text)
        if m:
            auction.level = int(m.group(1))
            auction.vocation = try_enum(Vocation, m.group(2).strip())
            auction.sex = try_enum(Sex, m.group(3).strip().lower())
            auction.world = m.group(4)
        outfit_img = auction_row.find("img", {"class": "AuctionOutfitImage"})
        m = id_addon_regex.search(outfit_img["src"])
        if m:
            auction.outfit = OutfitImage(image_url=outfit_img["src"], outfit_id=int(m.group(1)), addons=int(m.group(2)))
        item_boxes = auction_row.find_all("div", attrs={"class": "CVIcon"})
        for item_box in item_boxes:
            item = DisplayItem._parse_image_box(item_box)
            if item:
                auction.displayed_items.append(item)
        dates_containers = auction_row.find("div", {"class": "ShortAuctionData"})
        start_date_tag, end_date_tag, *_ = dates_containers.find_all("div", {"class": "ShortAuctionDataValue"})
        auction.auction_start = parse_tibia_datetime(start_date_tag.text.replace('\xa0', ' '))
        auction.auction_end = parse_tibia_datetime(end_date_tag.text.replace('\xa0', ' '))
        bids_container = auction_row.find("div", {"class": "ShortAuctionDataBidRow"})
        bid_tag = bids_container.find("div", {"class", "ShortAuctionDataValue"})
        bid_type_tag = bids_container.find_all("div", {"class", "ShortAuctionDataLabel"})[0]
        bid_type_str = bid_type_tag.text.replace(":", "").strip()
        auction.bid_type = try_enum(BidType, bid_type_str)
        auction.bid = parse_integer(bid_tag.text)
        auction_body_block = auction_row.find("div", {"class", "CurrentBid"})
        auction_info_tag = auction_body_block.find("div", {"class": "AuctionInfo"})
        status = ""
        if auction_info_tag:
            convert_line_breaks(auction_info_tag)
            status = auction_info_tag.text.replace("\n", " ").replace("  ", " ")
        auction.status = try_enum(AuctionStatus, status, AuctionStatus.IN_PROGRESS)
        argument_entries = auction_row.find_all("div", {"class": "Entry"})
        for entry in argument_entries:
            img = entry.find("img")
            img_url = img["src"]
            category_id = 0
            m = id_regex.search(img_url)
            if m:
                category_id = parse_integer(m.group(1))
            auction.sales_arguments.append(SalesArgument(content=entry.text, category_image=img_url,
                                                         category_id=category_id))
        return auction
    # endregion


class Auction(AuctionEntry):
    """The details of an auction.

    Attributes
    ----------
    auction_id: :class:`int`
        The internal id of the auction.
    name: :class:`str`
        The name of the character.
    level: :class:`int`
        The level of the character.
    world: :class:`str`
        The world the character is in.
    vocation: :class:`Vocation`
        The vocation of the character.
    sex: :class:`Sex`
        The sex of the character.
    outfit: :class:`OutfitImage`
        The current outfit selected by the user.
    displayed_items: :class:`list` of :class:`DisplayItem`
        The items selected to be displayed.
    sales_arguments: :class:`list` of :class:`SalesArgument`
        The sale arguments selected for the auction.
    auction_start: :class:`datetime.datetime`
        The date when the auction started.
    auction_end: :class:`datetime.datetime`
        The date when the auction ends.
    bid: :class:`int`
        The current bid in Tibia Coins.
    bid_type: :class:`BidType`
        The type of the auction's bid.
    status: :class:`AuctionStatus`
        The current status of the auction.
    hit_points: :class:`int`
        The hit points of the character.
    mana: :class:`int`
        The mana points of the character.
    capacity: :class:`int`
        The character's capacity in ounces.
    speed: :class:`int`
        The character's speed.
    blessings_count: :class:`int`
        The number of blessings the character has.
    outfits_count: :class:`int`
        The number of outfits the character has.
    titles_count: :class:`int`
        The number of titles the character has.
    skills: :class:`list` of :class:`SkillEntry`
        The current skills of the character.
    creation_date: :class:`datetime.datetime`
        The date when the character was created.
    experience: :class:`int`
        The total experience of the character.
    gold: :class:`int`
        The total amount of gold the character has.
    achievement_points: :class:`int`
        The number of achievement points of the character.
    regular_world_transfer_available_date: :class:`datetime.datetmie`
        The date after regular world transfers will be available to purchase and use.
        :obj:`None` indicates it is available immediately.
    charm_expansion: :class:`bool`
        Whether the character has a charm expansion or not.
    available_charm_points: :class:`int`
        The amount of charm points the character has available to spend.
    spent_charm_points: :class:`int`
        The total charm points the character has spent.
    prey_wildcards: :class:`int`
        The number of Prey Wildcards the character has.
    daly_reward_streak: :class:`int`
        The current daily reward streak.
    permanent_hunting_task_slots: :class:`int`
        The number of hunting task slots.
    permanent_prey_slots: :class:`int`
        The number of prey slots.
    hirelings: :class:`int`
        The number of hirelings the character has.
    hireling_jobs: :class:`int`
        The number of hireling jobs the character has.
    hireling_outfits: :class:`int`
        The number of hireling outfits the character has.
    items: :class:`ItemSummary`
        The items the character has across inventory, depot and item stash.
    store_items: :class:`ItemSummary`
        The store items the character has.
    mounts: :class:`Mounts`
        The mounts the character has unlocked.
    store_mounts: :class:`Mounts`
        The mounts the character has purchased from the store.
    outfits: :class:`Outfits`
        The outfits the character has unlocked.
    store_outfits: :class:`Outfits`
        The outfits the character has purchased from the store.
    familiars: :class:`Familiars`
        The familiars the character has purchased or unlocked.
    blessings: :class:`list` of :class:`BlessingEntry`
        The blessings the character has.
    imbuements: :class:`list` of :class:`str`
        The imbuements the character has unlocked access to.
    charms: :class:`list` of :class:`CharmEntry`
        The charms the character has unlocked.
    completed_cyclopedia_map_areas: :class:`list` of :class:`str`
        The cyclopedia map areas that the character has fully discovered.
    completed_quest_lines: :class:`list` of :class:`str`
        The quest lines the character has fully completed.
    titles: :class:`list` of :class:`str`
        The titles the character has unlocked.
    achievements: :class:`list` of :class:`AchievementEntry`
        The achievements the character has unlocked.
    bestiary_progress: :class:`list` of :class:`BestiaryEntry`
        The bestiary progress of the character.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.hit_points: int = kwargs.get("hit_points", 0)
        self.mana: int = kwargs.get("mana", 0)
        self.capacity: int = kwargs.get("capacity", 0)
        self.speed: int = kwargs.get("speed", 0)
        self.blessings_count: int = kwargs.get("blessings_count", 0)
        self.mounts_count: int = kwargs.get("mounts_count", 0)
        self.outfits_count: int = kwargs.get("outfits_count", 0)
        self.titles_count: int = kwargs.get("titles_count", 0)
        self.skills: List[SkillEntry] = kwargs.get("skills", [])
        self.creation_date: datetime.datetime = kwargs.get("creation_date")
        self.experience: int = kwargs.get("experience", 0)
        self.gold: int = kwargs.get("gold", 0)
        self.achievement_points: int = kwargs.get("achievement_points", 0)
        self.regular_world_transfer_available_date: datetime.datetime = kwargs.get(
            "regular_world_transfer_available_date",
        )
        self.charm_expansion: bool = kwargs.get("charm_expansion", False)
        self.available_charm_points: int = kwargs.get("available_charm_points", 0)
        self.spent_charm_points: int = kwargs.get("spent_charm_points", 0)
        self.daily_reward_streak: int = kwargs.get("daily_reward_streak", 0)
        self.hunting_task_points: int = kwargs.get("hunting_task_points", 0)
        self.prey_wildcards: int = kwargs.get("prey_wildcards", 0)
        self.permanent_hunting_task_slots: int = kwargs.get("permanent_hunting_task_slots", 0)
        self.permanent_prey_slots: int = kwargs.get("permanent_prey_slots", 0)
        self.hirelings: int = kwargs.get("hirelings", 0)
        self.hireling_jobs: int = kwargs.get("hireling_jobs", 0)
        self.hireling_outfits: int = kwargs.get("hireling_outfits", 0)
        self.items: ItemSummary = kwargs.get("items")
        self.store_items: ItemSummary = kwargs.get("store_items")
        self.mounts: Mounts = kwargs.get("mounts")
        self.store_mounts: Mounts = kwargs.get("store_mounts")
        self.outfits: Outfits = kwargs.get("outfits")
        self.store_outfits: Outfits = kwargs.get("store_outfits")
        self.blessings: List[BlessingEntry] = kwargs.get("blessings", [])
        self.imbuements: List[str] = kwargs.get("imbuements", [])
        self.charms: List[CharmEntry] = kwargs.get("charms", [])
        self.completed_cyclopedia_map_areas: List[str] = kwargs.get("completed_cyclopedia_map_areas", [])
        self.completed_quest_lines: List[str] = kwargs.get("completed_quest_lines", [])
        self.titles: List[str] = kwargs.get("titles", [])
        self.achievements: List[AchievementEntry] = kwargs.get("achievements", [])
        self.bestiary_progress: List[BestiaryEntry] = kwargs.get("bestiary_progress", [])

    __slots__ = (
        "hit_points",
        "mana",
        "capacity",
        "speed",
        "blessings_count",
        "mounts_count",
        "outfits_count",
        "titles_count",
        "skills",
        "creation_date",
        "experience",
        "gold",
        "achievement_points",
        "regular_world_transfer_available_date",
        "charm_expansion",
        "available_charm_points",
        "spent_charm_points",
        "daily_reward_streak",
        "hunting_task_points",
        "permanent_hunting_task_slots",
        "permanent_prey_slots",
        "prey_wildcards",
        "hirelings",
        "hireling_jobs",
        "hireling_outfits",
        "items",
        "store_items",
        "mounts",
        "store_mounts",
        "outfits",
        "store_outfits",
        "familiars",
        "blessings",
        "imbuements",
        "charms",
        "completed_cyclopedia_map_areas",
        "completed_quest_lines",
        "titles",
        "achievements",
        "bestiary_progress",
    )

    # region Properties
    @property
    def completed_bestiary_entries(self):
        """:class:`list` of :class:`BestiaryEntry`: Get a list of completed bestiary entries."""
        return [e for e in self.bestiary_progress if e.completed]

    @property
    def regular_world_transfer_available(self):
        """:class:`bool`: Whether regular world transfers are available immediately for this character."""
        return self.regular_world_transfer_available_date is None
    
    @property
    def skills_map(self) -> Dict[str, 'SkillEntry']:
        """:class:`dict` of :class:`str`, :class:`SkillEntry`: A mapping of skills by their name."""
        return {skill.name: skill for skill in self.skills}
    # endregion

    # region Public Methods
    @classmethod
    def from_content(cls, content, auction_id=0, skip_details=False):
        """Parse an auction detail page from Tibia.com and extracts its data.

        Parameters
        ----------
        content: :class:`str`
            The HTML content of the auction detail page in Tibia.com
        auction_id: :class:`int`, optional
            The ID of the auction.

            It is not possible to extract the ID from the page's content, so it may be passed to assign it manually.
        skip_details: :class:`bool`, optional
            Whether to skip parsing the entire auction and only parse the information shown in lists. False by default.

            This allows fetching basic information like name, level, vocation, world, bid and status, shaving off some
            parsing time.

        Returns
        -------
        :class:`Auction`
            The auction details if found, :obj:`None` otherwise.

        Raises
        ------
        InvalidContent
            If the content does not belong to a auction detail's page.
        """
        parsed_content = parse_tibiacom_content(content, builder='html5lib' if not skip_details else 'lxml')
        auction_row = parsed_content.find("div", attrs={"class": "Auction"})
        if not auction_row:
            if "internal error" in content:
                return None
            raise InvalidContent("content does not belong to a auction details page in Tibia.com")
        auction = cls._parse_auction(auction_row)
        auction.auction_id = auction_id
        if skip_details:
            return auction

        details_tables = cls._parse_tables(parsed_content)
        if "General" in details_tables:
            auction._parse_general_table(details_tables["General"])
        if "ItemSummary" in details_tables:
            auction.items = ItemSummary._parse_table(details_tables["ItemSummary"])
        if "StoreItemSummary" in details_tables:
            auction.store_items = ItemSummary._parse_table(details_tables["StoreItemSummary"])
        if "Mounts" in details_tables:
            auction.mounts = Mounts._parse_table(details_tables["Mounts"])
        if "StoreMounts" in details_tables:
            auction.store_mounts = Mounts._parse_table(details_tables["StoreMounts"])
        if "Outfits" in details_tables:
            auction.outfits = Outfits._parse_table(details_tables["Outfits"])
        if "StoreOutfits" in details_tables:
            auction.store_outfits = Outfits._parse_table(details_tables["StoreOutfits"])
        if "Familiars" in details_tables:
            auction.familiars = Familiars._parse_table(details_tables["Familiars"])
        if "Blessings" in details_tables:
            auction._parse_blessings_table(details_tables["Blessings"])
        if "Imbuements" in details_tables:
            auction.imbuements = cls._parse_single_column_table(details_tables["Imbuements"])
        if "Charms" in details_tables:
            auction._parse_charms_table(details_tables["Charms"])
        if "CompletedCyclopediaMapAreas" in details_tables:
            auction.completed_cyclopedia_map_areas = cls._parse_single_column_table(
                details_tables["CompletedCyclopediaMapAreas"])
        if "CompletedQuestLines" in details_tables:
            auction.completed_quest_lines = cls._parse_single_column_table(details_tables["CompletedQuestLines"])
        if "Titles" in details_tables:
            auction.titles = cls._parse_single_column_table(details_tables["Titles"])
        if "Achievements" in details_tables:
            auction._parse_achievements_table(details_tables["Achievements"])
        if "BestiaryProgress" in details_tables:
            auction._parse_bestiary_table(details_tables["BestiaryProgress"])
        return auction
    # endregion

    # region Private Methods
    @classmethod
    def _parse_tables(cls, parsed_content) -> Dict[str, bs4.Tag]:
        """Parse the character details tables.

        Parameters
        ----------
        parsed_content: :class:`bs4.Tag`
            The parsed content of the auction.

        Returns
        -------
        :class:`dict`
            A dictionary of the tables, grouped by their id.
        """
        details_tables = parsed_content.find_all("div", {"class": "CharacterDetailsBlock"})
        return {table["id"]: table for table in details_tables}

    @classmethod
    def _parse_data_table(cls, table) -> Dict[str, str]:
        """Parse a simple data table into a key value mapping.

        Parameters
        ----------
        table: :class:`bs4.Tag`
            The table to be parsed.

        Returns
        -------
        :class:`dict`
            A mapping containing the table's data.
        """
        rows = table.find_all("tr")
        data = {}
        for row in rows:
            name = row.find("span").text
            value = row.find("div").text
            name = name.lower().strip().replace(" ", "_").replace(":", "")
            data[name] = value
        return data

    def _parse_skills_table(self, table):
        """Parse the skills table.

        Parameters
        ----------
        table: :class:`bs4.Tag`
            The table containing the character's skill.
        """
        rows = table.find_all("tr")
        skills = []
        for row in rows:
            cols = row.find_all("td")
            name_c, level_c, progress_c = [c.text for c in cols]
            level = int(level_c)
            progress = float(progress_c.replace("%", ""))
            skills.append(SkillEntry(name=name_c, level=level, progress=progress))
        self.skills = skills

    def _parse_blessings_table(self, table):
        """Parse the blessings table.

        Parameters
        ----------
        table: :class:`bs4.Tag`
            The table containing the character's blessings.
        """
        table_content = table.find("table", attrs={"class": "TableContent"})
        _, *rows = table_content.find_all("tr")
        for row in rows:
            cols = row.find_all("td")
            amount_c, name_c = [c.text for c in cols]
            amount = int(amount_c.replace("x", ""))
            self.blessings.append(BlessingEntry(name_c, amount))

    @classmethod
    def _parse_single_column_table(cls, table):
        """Parse a table with a single column into an array.

        Parameters
        ----------
        table: :class:`bs4.Tag`
            A table with a single column.

        Returns
        -------
        :class:`list` of :class:`str`
            A list with the contents of each row.
        """
        table_content = table.find_all("table", attrs={"class": "TableContent"})[-1]
        _, *rows = table_content.find_all("tr")
        ret = []
        for row in rows:
            col = row.find("td")
            text = col.text
            if "more entries" in text:
                continue
            ret.append(text)
        return ret

    def _parse_charms_table(self, table):
        """Parse the charms table and extracts its information.

        Parameters
        ----------
        table: :class:`bs4.Tag`
            The table containing the charms.
        """
        table_content = table.find("table", attrs={"class": "TableContent"})
        _, *rows = table_content.find_all("tr")
        for row in rows:
            cols = row.find_all("td")
            if len(cols) != 2:
                continue
            cost_c, name_c = [c.text for c in cols]
            cost = parse_integer(cost_c.replace("x", ""))
            self.charms.append(CharmEntry(name_c, cost))

    def _parse_achievements_table(self, table):
        """Parse the achievements table and extracts its information.

        Parameters
        ----------
        table: :class:`bs4.Tag`
            The table containing the achievements.
        """
        table_content = table.find("table", attrs={"class": "TableContent"})
        _, *rows = table_content.find_all("tr")
        for row in rows:
            col = row.find("td")
            text = col.text.strip()
            if "more entries" in text:
                continue
            secret = col.find("img") is not None
            self.achievements.append(AchievementEntry(text, secret))

    def _parse_bestiary_table(self, table):
        """Parse the bestiary table and extracts its information.

        Parameters
        ----------
        table: :class:`bs4.Tag`
            The table containing the bestiary information.
        """
        table_content = table.find("table", attrs={"class": "TableContent"})
        _, *rows = table_content.find_all("tr")
        for row in rows:
            cols = row.find_all("td")
            if len(cols) != 3:
                continue
            step_c, kills_c, name_c = [c.text for c in cols]
            kills = parse_integer(kills_c.replace("x", ""))
            step = int(step_c)
            self.bestiary_progress.append(BestiaryEntry(name_c, kills, step))

    @classmethod
    def _parse_page_items(cls, content, entry_class):
        """Parse the elements of a page in the items, mounts and outfits.

        Attributes
        ----------
        content: :class:`str`
            The HTML content in the page.
        entry_class:
            The class defining the elements.

        Returns
        -------
            The entries contained in the page.
        """
        parsed_content = parse_tibiacom_content(content, builder='html5lib')
        item_boxes = parsed_content.find_all("div", attrs={"class": "CVIcon"})
        entries = []
        for item_box in item_boxes:
            item = entry_class._parse_image_box(item_box)
            if item:
                entries.append(item)
        return entries

    def _parse_general_table(self, table):
        """Parse the general information table and assigns its values.

        Parameters
        ----------
        table: :class:`bs4.Tag`
            The table with general information.
        """
        content_containers = table.find_all("table", {"class": "TableContent"})
        general_stats = self._parse_data_table(content_containers[0])
        self.hit_points = parse_integer(general_stats.get("hit_points", "0"))
        self.mana = parse_integer(general_stats.get("mana", "0"))
        self.capacity = parse_integer(general_stats.get("capacity", "0"))
        self.speed = parse_integer(general_stats.get("speed", "0"))
        self.mounts_count = parse_integer(general_stats.get("mounts", "0"))
        self.outfits_count = parse_integer(general_stats.get("outfits", "0"))
        self.titles_count = parse_integer(general_stats.get("titles", "0"))
        self.blessings_count = parse_integer(re.sub(r"/d+", "", general_stats.get("blessings", "0")))

        self._parse_skills_table(content_containers[1])

        additional_stats = self._parse_data_table(content_containers[2])
        self.creation_date = parse_tibia_datetime(additional_stats.get("creation_date", "").replace("\xa0", " "))
        self.experience = parse_integer(additional_stats.get("experience", "0"))
        self.gold = parse_integer(additional_stats.get("gold", "0"))
        self.achievement_points = parse_integer(additional_stats.get("achievement_points", "0"))

        transfer_data = self._parse_data_table(content_containers[3])
        transfer_text = transfer_data.get("regular_world_transfer")
        if "after" in transfer_text:
            date_string = transfer_text.split("after ")[1]
            self.regular_world_transfer_available_date = parse_tibia_datetime(date_string)

        charms_data = self._parse_data_table(content_containers[4])
        self.charm_expansion = "yes" in charms_data.get("charm_expansion", "")
        self.available_charm_points = parse_integer(charms_data.get("available_charm_points"))
        self.spent_charm_points = parse_integer(charms_data.get("spent_charm_points"))

        daily_rewards_data = self._parse_data_table(content_containers[5])
        self.daily_reward_streak = parse_integer(daily_rewards_data.popitem()[1])

        hunting_data = self._parse_data_table(content_containers[6])
        self.hunting_task_points = parse_integer(hunting_data.get("hunting_task_points", ""))
        self.permanent_hunting_task_slots = parse_integer(hunting_data.get("permanent_hunting_task_slots", ""))
        self.permanent_prey_slots = parse_integer(hunting_data.get("permanent_prey_slots", ""))
        self.prey_wildcards = parse_integer(hunting_data.get("prey_wildcards", ""))

        hirelings_data = self._parse_data_table(content_containers[7])
        self.hirelings = parse_integer(hirelings_data.get("hirelings", ""))
        self.hireling_jobs = parse_integer(hirelings_data.get("hireling_jobs", ""))
        self.hireling_outfits = parse_integer(hirelings_data.get("hireling_outfits", ""))
    # endregion


class OutfitImage(abc.Serializable):
    """The image of the outfit currently being worn by the character.

    Attributes
    ----------
    image_url: :class:`str`
        The URL of the image.
    outfit_id: :class:`int`
        The ID of the outfit.
    addons: :class:`int`
        The addons displayed in the outfit.
    """

    def __init__(self, **kwargs):
        self.image_url: str = kwargs.get("image_url")
        self.outfit_id: int = kwargs.get("outfit_id", 0)
        self.addons: int = kwargs.get("addons", 0)

    __slots__ = (
        "image_url",
        "outfit_id",
        "addons",
    )

    def __repr__(self):
        return f"<{self.__class__.__name__} outfit_id={self.outfit_id} addons={self.addons} " \
               f"image_url={self.image_url!r}>"


class PaginatedSummary(abc.Serializable):
    """Represents a paginated summary in the character auction section.

    Attributes
    ----------
    page: :class:`int`
        The current page being displayed.
    total_pages: :class:`int`
        The total number of pages.
    results: :class:`int`
        The total number of results.
    entries: :class:`list`
        The entries.
    fully_fetched: :class:`bool`
        Whether the summary was fetched completely, including all other pages.
    """

    entry_class = None

    def __init__(self, **kwargs):
        self.page: int = kwargs.get("page", 1)
        self.total_pages: int = kwargs.get("total_pages", 1)
        self.results: int = kwargs.get("results", 0)
        self.fully_fetched: bool = kwargs.get("fully_fetched", False)
        self.entries: List[DisplayImage] = kwargs.get("entries", [])

    __slots__ = (
        "page",
        "total_pages",
        "results",
        "fully_fetched",
        "entries",
    )

    def __repr__(self):
        return f"<{self.__class__.__name__} page={self.page} total_pages={self.total_pages} results={self.results} " \
               f"fully_fetched={self.fully_fetched} len(entries)={len(self.entries)}>"

    # region Public Methods
    def get_by_name(self, name):
        """Get an entry by its name.

        Parameters
        ----------
        name: :class:`str`
            The name of the entry, case insensitive.

        Returns
        -------
        :class:`object`:
            The entry matching the name.
        """
        return next((e for e in self.entries if e.name.lower() == name.lower()), None)

    def search(self, value):
        """Search an entry by its name.

        Parameters
        ----------
        value: :class:`str`
            The value to look for.

        Returns
        -------
        :class:`list`
            A list of entries with names containing the search term.
        """
        return [e for e in self.entries if value.lower() in e.name.lower()]

    def get_by_id(self, name):
        """Get an entry by its id.

        Parameters
        ----------
        name: :class:`str`
            The name of the entry, case insensitive.

        Returns
        -------
        :class:`object`:
            The entry matching the name.
        """
        return NotImplemented
    # endregion

    # region Private Methods
    def _parse_pagination(self, parsed_content):
        pagination_block = parsed_content.find("div", attrs={"class": "BlockPageNavigationRow"})
        if pagination_block is not None:
            self.page, self.total_pages, self.results = parse_pagination(pagination_block)
    # endregion


class ItemSummary(PaginatedSummary):
    """Items in a character's inventory and depot.

    Attributes
    ----------
    page: :class:`int`
        The current page being displayed.
    total_pages: :class:`int`
        The total number of pages.
    results: :class:`int`
        The total number of results.
    entries: :class:`list` of :class:`DisplayItem`
        The character's items.
    fully_fetched: :class:`bool`
        Whether the summary was fetched completely, including all other pages.
    """

    entries: List[DisplayItem]
    entry_class = DisplayItem

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_by_id(self, entry_id):
        """Get an item by its item id.

        Parameters
        ----------
        entry_id: :class:`int`
            The ID of the item.

        Returns
        -------
        :class:`DisplayItem`
            The item matching the id.
        """
        return next((e for e in self.entries if e.item_id == entry_id), None)

    @classmethod
    def _parse_table(cls, table):
        """Parse the item summary table.

        Parameters
        ----------
        table: :class:`bs4.Tag`
            The table containing the item summary.

        Returns
        -------
        :class:`ItemSummary`
            The item summary contained in the table.
        """
        summary = cls()
        summary._parse_pagination(table)
        item_boxes = table.find_all("div", attrs={"class": "CVIcon"})
        for item_box in item_boxes:
            item = DisplayItem._parse_image_box(item_box)
            if item:
                summary.entries.append(item)
        return summary


class Mounts(PaginatedSummary):
    """The mounts a character has unlocked or purchased.

    Attributes
    ----------
    page: :class:`int`
        The current page being displayed.
    total_pages: :class:`int`
        The total number of pages.
    results: :class:`int`
        The total number of results.
    entries: :class:`list` of :class:`DisplayMount`
        The character's mounts.
    fully_fetched: :class:`bool`
        Whether the summary was fetched completely, including all other pages.
    """

    entries: List[DisplayMount]
    entry_class = DisplayMount

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_by_id(self, entry_id):
        """Get a mount by its mount id.

        Parameters
        ----------
        entry_id: :class:`int`
            The ID of the mount.

        Returns
        -------
        :class:`DisplayMount`
            The mount matching the id.
        """
        return next((e for e in self.entries if e.mount_id == entry_id), None)

    @classmethod
    def _parse_table(cls, table):
        summary = cls()
        summary._parse_pagination(table)
        item_boxes = table.find_all("div", attrs={"class": "CVIcon"})
        for item_box in item_boxes:
            item = DisplayMount._parse_image_box(item_box)
            if item:
                summary.entries.append(item)
        return summary


class Familiars(PaginatedSummary):
    """The familiars the character has unlocked or purchased.

    Attributes
    ----------
    page: :class:`int`
        The current page being displayed.
    total_pages: :class:`int`
        The total number of pages.
    results: :class:`int`
        The total number of results.
    entries: :class:`list` of :class:`DisplayFamiliar`
        The familiars the character has unlocked or purchased.
    fully_fetched: :class:`bool`
        Whether the summary was fetched completely, including all other pages.
    """

    entries: List[DisplayFamiliar]
    entry_class = DisplayFamiliar

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_by_id(self, entry_id):
        """Get an outfit by its familiar id.

        Parameters
        ----------
        entry_id: :class:`int`
            The ID of the outfit.

        Returns
        -------
        :class:`DisplayOutfit`
            The outfit matching the id.
        """
        return next((e for e in self.entries if e.familiar_id == entry_id), None)

    @classmethod
    def _parse_table(cls, table):
        """Parse the outfits table.

        Parameters
        ----------
        table: :class:`bs4.Tag`
            The table containing the character outfits.

        Returns
        -------
        :class:`Outfits`
            The outfits contained in the table.
        """
        summary = cls()
        summary._parse_pagination(table)
        item_boxes = table.find_all("div", attrs={"class": "CVIcon"})
        for item_box in item_boxes:
            item = DisplayFamiliar._parse_image_box(item_box)
            if item:
                summary.entries.append(item)
        return summary


class Outfits(PaginatedSummary):
    """The outfits the character has unlocked or purchased.

    Attributes
    ----------
    page: :class:`int`
        The current page being displayed.
    total_pages: :class:`int`
        The total number of pages.
    results: :class:`int`
        The total number of results.
    entries: :class:`list` of :class:`DisplayOutfit`
        The outfits the character has unlocked or purchased.
    fully_fetched: :class:`bool`
        Whether the summary was fetched completely, including all other pages.
    """

    entries: List[DisplayOutfit]
    entry_class = DisplayOutfit

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_by_id(self, entry_id):
        """Get an outfit by its outfit id.

        Parameters
        ----------
        entry_id: :class:`int`
            The ID of the outfit.

        Returns
        -------
        :class:`DisplayOutfit`
            The outfit matching the id.
        """
        return next((e for e in self.entries if e.outfit_id == entry_id), None)

    @classmethod
    def _parse_table(cls, table):
        """Parse the outfits table.

        Parameters
        ----------
        table: :class:`bs4.Tag`
            The table containing the character outfits.

        Returns
        -------
        :class:`Outfits`
            The outfits contained in the table.
        """
        summary = cls()
        summary._parse_pagination(table)
        item_boxes = table.find_all("div", attrs={"class": "CVIcon"})
        for item_box in item_boxes:
            item = DisplayOutfit._parse_image_box(item_box)
            if item:
                summary.entries.append(item)
        return summary


class SalesArgument(abc.Serializable):
    """Represents a sales argument.

    Sales arguments can be selected when creating an auction, and allow the user to highlight certain
    character features in the auction listing.

    Attributes
    ----------
    category_image: :class:`str`
        The URL to the category icon.
    content: :class:`str`
        The content of the sales argument.
    """

    __slots__ = (
        "category_id",
        "category_image",
        "content",
    )

    def __init__(self, **kwargs):
        self.category_id: int = kwargs.get("category_id", 0)
        self.category_image: str = kwargs.get("category_image")
        self.content: str = kwargs.get("content")

    def __repr__(self):
        return f"<{self.__class__.__name__} category_id={self.category_id} content={self.content!r} " \
               f"category_image={self.category_image}>"


class SkillEntry(abc.Serializable):
    """Represents the character's skills.

    Attributes
    ----------
    name: :class:`name`
        The name of the skill.
    level: :class:`int`
        The current level.
    progress: :class:`float`
        The percentage of progress for the next level.
    """

    def __init__(self, **kwargs):
        self.name: str = kwargs.get("name")
        self.level: int = kwargs.get("level", 0)
        self.progress: float = kwargs.get("progress", 0.0)

    __slots__ = (
        "name",
        "level",
        "progress",
    )

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r} level={self.level} progress={self.progress}>"



import os
import re
import sys

from setuptools import find_packages, setup

if sys.version_info < (3, 7):
    sys.exit('Sorry, Python < 3.7 is not supported')


def get_version(package):
    """Return package version as listed in `__version__` in `init.py`."""
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


version = get_version("tibiapy")
if version.endswith(('a', 'b', 'rc')):
    # append version identifier based on commit count
    try:
        import subprocess
        p = subprocess.Popen(['git', 'rev-list', '--count', 'HEAD'],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = p.communicate()
        if out:
            version += out.decode('utf-8').strip()
        p = subprocess.Popen(['git', 'rev-parse', '--short', 'HEAD'],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = p.communicate()
        if out:
            version += '+g' + out.decode('utf-8').strip()
    except Exception:
        pass

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

with open('README.md') as f:
    readme = f.read()

extras_require = {
    'docs': [
        'sphinx',
    ],
    'test': [
        'asynctest',
        'aioresponses',
        'coverage',
    ],
}

setup(
    name='tibia.py',
    version=version,
    author='Galarzaa90',
    author_email="allan.galarza@gmail.com",
    url='https://github.com/Galarzaa90/tibia.py',
    license='Apache 2.0',
    install_requires=requirements,
    extras_require=extras_require,
    description="API that parses website content into python data.",
    long_description=readme,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    include_package_data=True,
    project_urls={
        "GitHub: Repo": "https://github.com/Galarzaa90/tibia.py",
        "GitHub: Issues": "https://github.com/Galarzaa90/tibia.py/issues",
        "Docs: RTD": "https://tibiapy.readthedocs.io/en/stable/",
        "Docs: Changelog": "https://tibiapy.readthedocs.io/en/stable/changelog.html",
        "Coverage: Codecov": "https://app.codecov.io/gh/Galarzaa90/tibia.py",
        "Docker Hub: Repo": "https://hub.docker.com/repository/docker/galarzaa90/tibia.py",
        "SonarCloud": "https://sonarcloud.io/dashboard?id=Galarzaa90_tibia.py",
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Games/Entertainment :: Role-Playing',
        'Topic :: Internet',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Text Processing :: Markup :: HTML',
        'Topic :: Utilities',
        'Typing :: Typed',
    ],
)



"""Models related to the guilds section in Tibia.com."""
import datetime
import re
from collections import OrderedDict, defaultdict
from typing import Dict, List, Optional

import bs4

from tibiapy import abc
from tibiapy.enums import Vocation
from tibiapy.errors import InvalidContent
from tibiapy.house import GuildHouse
from tibiapy.utils import (get_tibia_url, parse_form_data, parse_tibia_date, parse_tibiacom_content, try_date,
                           try_datetime, try_enum)

__all__ = (
    "Guild",
    "GuildMember",
    "GuildInvite",
    "GuildsSection",
    "GuildWars",
    "GuildWarEntry",
    "GuildEntry",
)

COLS_INVITED_MEMBER = 2
COLS_GUILD_MEMBER = 6

founded_regex = re.compile(
    r'(?P<desc>.*)The guild was founded on (?P<world>\w+) on (?P<date>[^.]+)\.\nIt is (?P<status>[^.]+).', re.DOTALL)
applications_regex = re.compile(r'Guild is (\w+) for applications\.')
homepage_regex = re.compile(r'The official homepage is at ([\w.]+)\.')
guildhall_regex = re.compile(r'Their home on \w+ is (?P<name>[^.]+). The rent is paid until (?P<date>[^.]+)')
disband_regex = re.compile(r'It will be disbanded on (\w+\s\d+\s\d+)\s([^.]+).')
disband_tibadata_regex = re.compile(r'It will be disbanded, ([^.]+).')
title_regex = re.compile(r'([^(]+)\(([^)]+)\)')

war_guilds_regegx = re.compile(r'The guild ([\w\s]+) is at war with the guild ([^.]+).')
war_score_regex = re.compile(r'scored ([\d,]+) kills? against')
war_fee_regex = re.compile(r'the guild [\w\s]+ wins the war, they will receive ([\d,]+) gold.')
war_score_limit_regex = re.compile(r'guild scores ([\d,]+) kills against')
war_end_regex = re.compile(r'war will end on (\w{3}\s\d{2}\s\d{4})')

war_history_header_regex = re.compile(r'guild ([\w\s]+) fought against ([\w\s]+).')
war_start_duration_regex = re.compile(r'started on (\w{3}\s\d{2}\s\d{4}) and had been set for a duration of (\w+) days')
kills_needed_regex = re.compile(r'(\w+) kills were needed')
war_history_fee_regex = re.compile(r'agreed on a fee of (\w+) gold for the guild [\w\s]+ and a fee of (\d+) gold')
surrender_regex = re.compile(r'(?:The guild ([\w\s]+)|A disbanded guild) surrendered on (\w{3}\s\d{2}\s\d{4})')
war_ended_regex = re.compile(r'war ended on (\w{3}\s\d{2}\s\d{4}) when the guild ([\w\s]+) had reached the')
war_score_end_regex = re.compile(r'scored (\d+) kills against')

war_current_empty = re.compile(r'The guild ([\w\s]+) is currently not')


class GuildsSection(abc.Serializable):
    """The guilds section in Tibia.com.

    .. versionadded:: 5.0.0

    Attributes
    ----------
    world: :class:`str`
        The name of the world. If :obj:`None`, the section belongs to a world that doesn't exist.
    entries: :class:`list` of :class:`GuildEntry`
        The list of guilds in the world.
    available_worlds: :class:`list` of :class:`str`
        The list of worlds available for selection.
    """

    __slots__ = (
        "world",
        "entries",
        "available_worlds",
    )

    def __init__(self, world, entries=None, available_worlds=None):
        self.world: str = world
        self.entries: List[GuildEntry] = entries or []
        self.available_worlds: List[str] = available_worlds or []

    def __repr__(self):
        return f"<{self.__class__.__name__} world={self.world!r} len(entries)={len(self.entries)}>"

    @property
    def active_guilds(self):
        """:class:`list` of :class:`GuildEntry`: Get a list of the guilds that are active."""
        return [g for g in self.entries if g.active]

    @property
    def in_formation_guilds(self):
        """:class:`list` of :class:`GuildEntry`: Get a list of the guilds that are in course of formation."""
        return [g for g in self.entries if not g.active]

    @property
    def url(self):
        """:class:`str`: Get the URL to this guild section."""
        return self.get_url(self.world)

    @classmethod
    def from_content(cls, content):
        """Get a list of guilds from the HTML content of the world guilds' page.

        Parameters
        ----------
        content: :class:`str`
            The HTML content of the page.

        Returns
        -------
        :class:`GuildsSection`
            List of guilds in the current world. :obj:`None` if it's the list of a world that doesn't exist.

        Raises
        ------
        InvalidContent
            If content is not the HTML of a guild's page.
        """
        try:
            parsed_content = parse_tibiacom_content(content)
            form = parsed_content.find("form")
            data = parse_form_data(form, include_options=True)
            selected_world = data["world"] if data["world"] else None
            available_worlds = [w for w in data["__options__"]["world"].values() if w]
            guilds = cls(selected_world, available_worlds=available_worlds)
        except AttributeError as e:
            raise InvalidContent("Content does not belong to world guild list.", e)
        # First TableContainer contains world selector.
        _, *containers = parsed_content.find_all('div', class_="TableContainer")
        for container in containers:
            header = container.find('div', class_="Text")
            active = "Active" in header.text
            header, *rows = container.find_all("tr", {'bgcolor': ["#D4C0A1", "#F1E0C6"]})
            for row in rows:
                columns = row.find_all('td')
                logo_img = columns[0].find('img')["src"]
                description_lines = columns[1].get_text("\n").split("\n", 1)
                name = description_lines[0]
                description = None
                if len(description_lines) > 1:
                    description = description_lines[1].replace("\r", "").replace("\n", " ")
                guild = GuildEntry(name, guilds.world, logo_img, description, active)
                guilds.entries.append(guild)
        return guilds

    @classmethod
    def get_url(cls, world):
        """Get the Tibia.com URL for the guild section of a specific world.

        Parameters
        ----------
        world: :class:`str`
            The name of the world.

        Returns
        -------
        :class:`str`
            The URL to the guild's page
        """
        return get_tibia_url("community", "guilds", world=world)


class Guild(abc.BaseGuild, abc.Serializable):
    """A Tibia guild, viewed from its guild's page.

    Attributes
    ------------
    name: :class:`str`
        The name of the guild.
    logo_url: :class:`str`
        The URL to the guild's logo.
    description: :class:`str`, optional
        The description of the guild.
    world: :class:`str`
        The world this guild belongs to.
    founded: :class:`datetime.date`
        The day the guild was founded.
    active: :class:`bool`
        Whether the guild is active or still in formation.
    guildhall: :class:`GuildHouse`, optional
        The guild's guildhall if any.
    open_applications: :class:`bool`
        Whether applications are open or not.
    active_war: :class:`bool`
        Whether the guild is currently in an active war or not.

        .. versionadded:: 3.0.0
    disband_date: :class:`datetime.datetime`, optional
        The date when the guild will be disbanded if the condition hasn't been meet.
    disband_condition: :class:`str`, optional
        The reason why the guild will get disbanded.
    homepage: :class:`str`, optional
        The guild's homepage, if any.
    members: :class:`list` of :class:`GuildMember`
        List of guild members.
    invites: :class:`list` of :class:`GuildInvite`
        List of invited characters.
    """

    __slots__ = (
        "world",
        "logo_url",
        "description",
        "founded",
        "active",
        "guildhall",
        "open_applications",
        "active_war",
        "disband_condition",
        "disband_date",
        "homepage",
        "members",
        "invites",
    )

    _serializable_properties = (
        "member_count",
        "online_count",
        "ranks",
    )

    def __init__(self, name=None, world=None, **kwargs):
        self.name: str = name
        self.world: str = world
        self.logo_url: str = kwargs.get("logo_url")
        self.description: Optional[str] = kwargs.get("description")
        self.founded = try_date(kwargs.get("founded"))
        self.active: bool = kwargs.get("active", False)
        self.guildhall: Optional[GuildHouse] = kwargs.get("guildhall")
        self.open_applications: bool = kwargs.get("open_applications", False)
        self.active_war: bool = kwargs.get("active_war", False)
        self.disband_condition: Optional[str] = kwargs.get("disband_condition")
        self.disband_date = try_datetime(kwargs.get("disband_date"))
        self.homepage: Optional[str] = kwargs.get("homepage")
        self.members: List[GuildMember] = kwargs.get("members", [])
        self.invites: List[GuildInvite] = kwargs.get("invites", [])

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r} world={self.world!r}>"

    # region Properties
    @property
    def member_count(self):
        """:class:`int`: The number of members in the guild."""
        return len(self.members)

    @property
    def online_count(self):
        """:class:`int`: The number of online members in the guild."""
        return len(self.online_members)

    @property
    def online_members(self):
        """:class:`list` of :class:`GuildMember`: List of currently online members."""
        return list(filter(lambda m: m.online, self.members))

    @property
    def ranks(self) -> List[str]:
        """:class:`list` of :class:`str`: Ranks in their hierarchical order."""
        return list(OrderedDict.fromkeys((m.rank for m in self.members)))

    @property
    def members_by_rank(self) -> Dict[str, List['GuildMember']]:
        """:class:`dict`: Get a mapping of members, grouped by their guild rank."""
        rank_dict = defaultdict(list)
        [rank_dict[m.rank].append(m) for m in self.members]
        return dict(rank_dict)
    # endregion

    # region Public methods
    @classmethod
    def from_content(cls, content):
        """Create an instance of the class from the HTML content of the guild's page.

        Parameters
        -----------
        content: :class:`str`
            The HTML content of the page.

        Returns
        ----------
        :class:`Guild`
            The guild contained in the page or None if it doesn't exist.

        Raises
        ------
        InvalidContent
            If content is not the HTML of a guild's page.
        """
        if "An internal error has occurred" in content:
            return None

        parsed_content = parse_tibiacom_content(content)
        try:
            name_header = parsed_content.find('h1')
            guild = Guild(name_header.text.strip())
        except AttributeError:
            raise InvalidContent("content does not belong to a Tibia.com guild page.")

        if not guild._parse_logo(parsed_content):
            raise InvalidContent("content does not belong to a Tibia.com guild page.")

        info_container = parsed_content.find("div", id="GuildInformationContainer")
        guild._parse_guild_info(info_container)
        guild._parse_application_info(info_container)
        guild._parse_guild_homepage(info_container)
        guild._parse_guild_guildhall(info_container)
        guild._parse_guild_disband_info(info_container)
        guild._parse_guild_members(parsed_content)

        if guild.guildhall and guild.members:
            guild.guildhall.owner = guild.members[0].name

        return guild

    # endregion

    # region Private methods
    def _parse_current_member(self, previous_rank, values):
        """Parse the column texts of a member row into a member dictionary.

        Parameters
        ----------
        previous_rank: :class:`dict`[int, str]
            The last rank present in the rows.
        values: :class:`tuple` of :class:`str`
            A list of row contents.
        """
        rank, name, vocation, level, joined, status = values
        rank = previous_rank[1] if rank == " " else rank
        title = None
        previous_rank[1] = rank
        m = title_regex.match(name)
        if m:
            name = m.group(1)
            title = m.group(2)
        self.members.append(GuildMember(name.strip(), rank.strip(), title, int(level), vocation, joined=joined,
                                        online=status == "online"))

    def _parse_application_info(self, info_container):
        """
        Parse the guild's application info.

        Parameters
        ----------
        info_container: :class:`bs4.Tag`
            The parsed content of the information container.
        """
        m = applications_regex.search(info_container.text)
        if m:
            self.open_applications = m.group(1) == "opened"
        self.active_war = "during war" in info_container.text

    def _parse_guild_disband_info(self, info_container):
        """
        Parse the guild's disband info, if available.

        Parameters
        ----------
        info_container: :class:`bs4.Tag`
            The parsed content of the information container.
        """
        m = disband_regex.search(info_container.text)
        if m:
            self.disband_condition = m.group(2)
            self.disband_date = parse_tibia_date(m.group(1).replace("\xa0", " "))

    def _parse_guild_guildhall(self, info_container):
        """
        Parse the guild's guildhall info.

        Parameters
        ----------
        info_container: :class:`bs4.Tag`
            The parsed content of the information container.
        """
        m = guildhall_regex.search(info_container.text)
        if m:
            paid_until = parse_tibia_date(m.group("date").replace("\xa0", " "))
            self.guildhall = GuildHouse(m.group("name"), self.world, paid_until_date=paid_until)

    def _parse_guild_homepage(self, info_container):
        """Parse the guild's homepage info.

        Parameters
        ----------
        info_container: :class:`bs4.Tag`
            The parsed content of the information container.
        """
        m = homepage_regex.search(info_container.text)
        if m:
            self.homepage = m.group(1)

    def _parse_guild_info(self, info_container):
        """
        Parse the guild's general information and applies the found values.

        Parameters
        ----------
        info_container: :class:`bs4.Tag`
            The parsed content of the information container.
        """
        m = founded_regex.search(info_container.text)
        if m:
            description = m.group("desc").strip()
            self.description = description if description else None
            self.world = m.group("world")
            self.founded = parse_tibia_date(m.group("date").replace("\xa0", " "))
            self.active = "currently active" in m.group("status")

    def _parse_logo(self, parsed_content):
        """Parse the guild logo and saves it to the instance.

        Parameters
        ----------
        parsed_content: :class:`bs4.Tag`
            The parsed content of the page.

        Returns
        -------
        :class:`bool`
            Whether the logo was found or not.
        """
        logo_img = parsed_content.find('img', {'height': '64'})
        if logo_img is None:
            return False

        self.logo_url = logo_img["src"]
        return True

    def _parse_guild_members(self, parsed_content):
        """
        Parse the guild's member and invited list.

        Parameters
        ----------
        parsed_content: :class:`bs4.Tag`
            The parsed content of the guild's page
        """
        member_rows = parsed_content.find_all("tr", {'bgcolor': ["#D4C0A1", "#F1E0C6"]})
        previous_rank = {}
        for row in member_rows:
            columns = row.find_all('td')
            values = tuple(c.text.replace("\u00a0", " ") for c in columns)
            if len(columns) == COLS_GUILD_MEMBER:
                self._parse_current_member(previous_rank, values)
            if len(columns) == COLS_INVITED_MEMBER:
                self._parse_invited_member(values)

    def _parse_invited_member(self, values):
        """Parse the column texts of an invited row into a invited dictionary.

        Parameters
        ----------
        values: tuple[:class:`str`]
            A list of row contents.
        """
        name, date = values
        if date != "Invitation Date":
            self.invites.append(GuildInvite(name, date))
    # endregion


class GuildMember(abc.BaseCharacter, abc.Serializable):
    """Represents a guild member.

    Attributes
    --------------
    rank: :class:`str`
        The rank the member belongs to
    name: :class:`str`
        The name of the guild member.
    title: :class:`str`, optional
        The member's title.
    level: :class:`int`
        The member's level.
    vocation: :class:`Vocation`
        The member's vocation.
    joined: :class:`datetime.date`
        The day the member joined the guild.
    online: :class:`bool`
        Whether the member is online or not.
    """

    __slots__ = (
        "name",
        "rank",
        "title",
        "level",
        "vocation",
        "joined",
        "online",
    )

    def __init__(self, name=None, rank=None, title=None, level=0, vocation=None, **kwargs):
        self.name: str = name
        self.rank: str = rank
        self.title: Optional[str] = title
        self.vocation: Vocation = try_enum(Vocation, vocation)
        self.level = int(level)
        self.online: bool = kwargs.get("online", False)
        self.joined: datetime.date = try_date(kwargs.get("joined"))

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r} rank={self.rank!r} level={self.level} " \
               f"vocation={self.vocation!r}>"


class GuildInvite(abc.BaseCharacter, abc.Serializable):
    """Represents an invited character.

    Attributes
    ------------
    name: :class:`str`
        The name of the character

    date: :class:`datetime.date`
        The day when the character was invited.
    """

    __slots__ = (
        "name",
        "date",
    )

    def __init__(self, name=None, date=None):
        self.name: str = name
        self.date: datetime.date = try_date(date)

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r} date={self.date!r}>"


class GuildWars(abc.Serializable):
    """Represents a guild's wars.

    .. versionadded:: 3.0.0

    Attributes
    ----------
    name: :class:`str`
        The name of the guild.
    current: :class:`GuildWarEntry`
        The current war the guild is involved in.
    history: :class:`list` of :class:`GuildWarEntry`
        The previous wars the guild has been involved in.
    """

    __slots__ = (
        'name',
        'current',
        'history',
    )

    def __init__(self, name, current=None, history=None):
        self.name: str = name
        self.current: Optional[GuildWarEntry] = current
        self.history: List[GuildWarEntry] = history or []

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r}>"

    @property
    def url(self):
        """:class:`str`: The URL of this guild's war page on Tibia.com."""
        return self.get_url(self.name)

    @classmethod
    def get_url(cls, name):
        """Get the URL to the guild's war page of a guild with the given name.

        Parameters
        ----------
        name: :class:`str`
            The name of the guild.

        Returns
        -------
        :class:`str`
            The URL to the guild's war page.
        """
        return Guild.get_url_wars(name)

    @classmethod
    def from_content(cls, content):
        """Get a guild's war information from Tibia.com's content.

        Parameters
        ----------
        content: :class:`str`
            The HTML content of a guild's war section in Tibia.com

        Returns
        -------
        :class:`GuildWars`
            The guild's war information.
        """
        try:
            parsed_content = parse_tibiacom_content(content)
            table_current, table_history = parsed_content.find_all("div", attrs={"class": "TableContainer"})
            current_table_content = table_current.find("table", attrs={"class": "TableContent"})
            current_war = None
            guild_name = None
            if current_table_content is not None:
                for br in current_table_content.find_all("br"):
                    br.replace_with("\n")
                current_war = cls._parse_current_war_information(current_table_content.text)
            else:
                current_war_text = table_current.text
                current_war_match = war_current_empty.search(current_war_text)
                guild_name = current_war_match.group(1)

            history_entries = []
            history_contents = table_history.find_all("table", attrs={"class": "TableContent"})
            for history_content in history_contents:
                for br in history_content.find_all("br"):
                    br.replace_with("\n")
                entry = cls._parse_war_history_entry(history_content.text)
                history_entries.append(entry)

            if current_war:
                guild_name = current_war.guild_name
            elif history_entries:
                guild_name = history_entries[0].guild_name

            return cls(guild_name, current=current_war, history=history_entries)
        except ValueError as e:
            raise InvalidContent("content does not belong to the guild wars section", e)

    @classmethod
    def _parse_current_war_information(cls, text):
        """Parse the guild's current war information.

        Parameters
        ----------
        text: :class:`str`
            The text describing the current war's information.

        Returns
        -------
        :class:`GuildWarEntry`
            The guild's war entry for the current war.
        """
        text = text.replace('\xa0', ' ').strip()
        names_match = war_guilds_regegx.search(text)
        guild_name, opposing_name = names_match.groups()
        scores_match = war_score_regex.findall(text)
        guild_score, opposing_score = scores_match
        fee_match = war_fee_regex.findall(text)
        guild_fee, opposing_fee = fee_match

        score_limit_match = war_score_limit_regex.search(text)
        score_limit = score_limit_match.group(1)

        end_date_match = war_end_regex.search(text)
        end_date_str = end_date_match.group(1)
        end_date = parse_tibia_date(end_date_str)

        return GuildWarEntry(guild_name=guild_name, opponent_name=opposing_name, guild_score=int(guild_score),
                             opponent_score=int(opposing_score), guild_fee=int(guild_fee),
                             opponent_fee=int(opposing_fee), score_limit=int(score_limit), end_date=end_date)

    @classmethod
    def _parse_war_history_entry(cls, text):
        """Parse a guild's war information.

        Parameters
        ----------
        text: :class:`str`
            The text describing the war's information.

        Returns
        -------
        :class:`GuildWarEntry`
            The guild's war entry described in the text..
        """
        text = text.replace('\xa0', ' ').strip()
        header_match = war_history_header_regex.search(text)
        guild_name, opposing_name = header_match.groups()
        if "disbanded guild" in opposing_name:
            opposing_name = None
        start_duration_match = war_start_duration_regex.search(text)
        start_str, duration_str = start_duration_match.groups()
        start_date = parse_tibia_date(start_str)
        duration = datetime.timedelta(days=int(duration_str))
        kills_match = kills_needed_regex.search(text)
        kills_needed = int(kills_match.group(1))
        fee_match = war_history_fee_regex.search(text)
        guild_fee, opponent_fee = fee_match.groups()
        winner = None
        surrender = False
        end_date = None
        guild_score = opponent_score = 0
        surrender_match = surrender_regex.search(text)
        if surrender_match:
            surrending_guild = surrender_match.group(1)
            end_date = parse_tibia_date(surrender_match.group(2))
            winner = guild_name if surrending_guild != guild_name else opposing_name
            surrender = True

        war_score_match = war_score_regex.findall(text)
        if war_score_match and len(war_score_match) == 2:
            guild_score, opponent_score = war_score_match
            guild_score = int(guild_score)
            opponent_score = int(guild_score)

        war_end_match = war_ended_regex.search(text)
        if war_end_match:
            end_date = parse_tibia_date(war_end_match.group(1))
            winning_guild = war_end_match.group(2)
            if "disbanded guild" in winning_guild:
                winning_guild = None
            winner = guild_name if winning_guild == guild_name else opposing_name
            loser_score_match = war_score_end_regex.search(text)
            loser_score = int(loser_score_match.group(1)) if loser_score_match else 0
            guild_score = kills_needed if guild_name == winner else loser_score
            opponent_score = kills_needed if guild_name != winner else loser_score

        if "no guild had reached the needed kills" in text:
            winner = guild_name if guild_score > opponent_score else opposing_name

        return GuildWarEntry(guild_name=guild_name, opponent_name=opposing_name, start_date=start_date,
                             duration=duration, score_limit=kills_needed, guild_fee=int(guild_fee),
                             opponent_fee=int(opponent_fee), surrender=surrender, winner=winner, end_date=end_date,
                             opponent_score=opponent_score, guild_score=guild_score)


class GuildWarEntry(abc.Serializable):
    """Represents a guild war entry.

    .. versionadded:: 3.0.0

    Attributes
    ----------
    guild_name: :class:`str`
        The name of the guild.
    guild_score: :class:`int`
        The number of kills the guild has scored.
    guild_fee: :class:`int`
        The number of gold coins the guild will pay if they lose the war.
    opponent_name: :class:`str`
        The name of the opposing guild. If the guild no longer exist, this will be :obj:`None`.
    opponent_score: :class:`int`
        The number of kills the opposing guild has scored.
    opponent_fee: :class:`int`
        The number of gold coins the opposing guild will pay if they lose the war.
    start_date: :class:`datetime.date`
        The date when the war started.

        When a war is in progress, the start date is not visible.
    score_limit: :class:`int`
        The number of kills needed to win the war.
    duration: :class:`datetime.timedelta`
        The set duration of the war.

        When a war is in progress, the duration is not visible.
    end_date: :class:`datetime.date`
        The deadline for the war to finish if the score is not reached for wars in progress, or the date when the
        war ended.
    winner: :class:`str`
        The name of the guild that won.

        Note that if the winning guild is disbanded, this may be :obj:`None`.
    surrender: :class:`bool`
        Whether the losing guild surrendered or not.
    """

    __slots__ = (
        "guild_name",
        "guild_score",
        "guild_fee",
        "opponent_name",
        "opponent_score",
        "opponent_fee",
        "start_date",
        "score_limit",
        "duration",
        "end_date",
        "winner",
        "surrender",
    )

    def __init__(self, **kwargs):
        self.guild_name: str = kwargs.get("guild_name")
        self.guild_score: int = kwargs.get("guild_score", 0)
        self.guild_fee: int = kwargs.get("guild_fee", 0)
        self.opponent_name: Optional[str] = kwargs.get("opponent_name")
        self.opponent_score: str = kwargs.get("opponent_score", 0)
        self.opponent_fee: int = kwargs.get("opponent_fee", 0)
        self.start_date: datetime.datetime = kwargs.get("start_date")
        self.score_limit: int = kwargs.get("score_limit", 0)
        self.duration: datetime.timedelta = kwargs.get("duration")
        self.end_date: datetime.datetime = kwargs.get("end_date")
        self.winner: Optional[str] = kwargs.get("winner")
        self.surrender: bool = kwargs.get("surrender", False)

    def __repr__(self):
        return f"<{self.__class__.__name__} guild_name={self.guild_name!r} opponent_name={self.opponent_name!r}>"

    @property
    def guild_url(self):
        """:class:`str`: The URL to the guild's information page on Tibia.com."""
        return Guild.get_url(self.guild_name)

    @property
    def opponent_guild_url(self):
        """:class:`str`: The URL to the opposing guild's information page on Tibia.com."""
        return Guild.get_url(self.opponent_name) if self.opponent_name else None


class GuildEntry(abc.BaseGuild, abc.Serializable):
    """Represents a Tibia guild in the guild list of a world.

    Attributes
    ------------
    name: :class:`str`
        The name of the guild.
    logo_url: :class:`str`
        The URL to the guild's logo.
    description: :class:`str`, optional
        The description of the guild.
    world: :class:`str`
        The world this guild belongs to.
    active: :class:`bool`
        Whether the guild is active or still in formation.
    """

    __slots__ = (
        "name",
        "logo_url",
        "description",
        "world",
        "active",
    )

    def __init__(self, name, world, logo_url=None, description=None, active=False):
        self.name: str = name
        self.world: str = world
        self.logo_url: str = logo_url
        self.description: Optional[str] = description
        self.active: bool = active



import unittest

from tests.tests_tibiapy import TestCommons
from tibiapy import InvalidContent, CreaturesSection, Creature, CreatureEntry

FILE_CREATURE_SECTION = "library/creature_list.txt"
FILE_CREATURE_CONVINCEABLE = "library/creature_convinceable.txt"
FILE_CREATURE_ELEMENTAL_RESISTANCES = "library/creature_elemental_resistances.txt"


class TestCreature(TestCommons, unittest.TestCase):
    # region Tibia.com Tests
    def test_creatures_section_from_boosted_creature_header_content(self):
        """Testing parsing the boosted creture from any tibia.com page."""
        content = self.load_resource(self.FILE_UNRELATED_SECTION)
        creature = CreaturesSection.boosted_creature_from_header(content)

        self.assertIsInstance(creature, CreatureEntry)
        self.assertEqual("Skeleton Warrior", creature.name)

    def test_creatures_section_from_boosted_creature_header_content_not_tibiacom(self):
        """Testing parsing the boosted creature from a page that is not Tibia.com"""
        with self.assertRaises(InvalidContent):
            CreaturesSection.boosted_creature_from_header("<html><div><p>Nothing</p></div></html>")

    def test_creature_section_from_content(self):
        """Test parsing the creatures section"""
        content = self.load_resource(FILE_CREATURE_SECTION)
        creatures = CreaturesSection.from_content(content)

        self.assertIsNotNone(creatures)
        self.assertIsNotNone(creatures.boosted_creature)
        self.assertEqual("Blood Crab", creatures.boosted_creature.name)
        self.assertEqual("bloodcrab", creatures.boosted_creature.identifier)
        self.assertIsNotNone(creatures.boosted_creature.image_url)
        self.assertIsNotNone(creatures.boosted_creature.url)
        self.assertEqual(536, len(creatures.creatures))
        for creature in creatures.creatures:
            with self.subTest(name=creature.name):
                self.assertIsInstance(creature.name, str)
                self.assertIsInstance(creature.identifier, str)

    def test_creatures_section_from_content_invalid_content(self):
        """Testing parsing the creatures section from an invalid section"""
        content = self.load_resource(self.FILE_UNRELATED_SECTION)
        with self.assertRaises(InvalidContent):
            CreaturesSection.from_content(content)

    def test_creature_from_content(self):
        content = self.load_resource(FILE_CREATURE_CONVINCEABLE)
        creature = Creature.from_content(content)

        self.assertIsNotNone(creature)
        self.assertEqual("Fish", creature.name)
        self.assertEqual("fish", creature.identifier)
        self.assertEqual(25, creature.hitpoints)
        self.assertEqual(0, creature.experience)
        self.assertEqual(0, creature.mana_cost)
        self.assertFalse(creature.summonable)
        self.assertTrue(creature.convinceable)
        self.assertEqual("nothing", creature.loot)

    def test_creature_from_content_elemental_resistances(self):
        content = self.load_resource(FILE_CREATURE_ELEMENTAL_RESISTANCES)
        creature = Creature.from_content(content)

        self.assertIsNotNone(creature)
        self.assertEqual("Dragons", creature.name)
        self.assertEqual("dragon", creature.identifier)
        self.assertEqual(1000, creature.hitpoints)
        self.assertEqual(700, creature.experience)
        self.assertIsNone(creature.mana_cost)
        self.assertFalse(creature.summonable)
        self.assertFalse(creature.convinceable)
        self.assertIsNotNone(creature.description)
        self.assertIn("energy", creature.strong_against)
        self.assertIn("fire", creature.immune_to)
        self.assertIn("ice", creature.weak_against)

    def test_creature_from_content_invalid_content(self):
        """Testing parsing the creatures section from an invalid section"""
        content = self.load_resource(self.FILE_UNRELATED_SECTION)

        creature = Creature.from_content(content)

        self.assertIsNone(creature)
    # endregion



"""Models related to the highscores section in Tibia.com."""
import datetime
import re
from collections import OrderedDict
from typing import List, Optional

from tibiapy import abc
from tibiapy.enums import Category, Vocation, VocationFilter, BattlEyeTypeFilter, PvpTypeFilter, \
    BattlEyeHighscoresFilter
from tibiapy.errors import InvalidContent
from tibiapy.utils import get_tibia_url, parse_form_data, parse_tibiacom_content, try_enum, parse_integer

__all__ = (
    "Highscores",
    "HighscoresEntry",
    "LoyaltyHighscoresEntry",
)

results_pattern = re.compile(r'Results: ([\d,]+)')
numeric_pattern = re.compile(r'(\d+)')


class Highscores(abc.Serializable):
    """Represents the highscores of a world.

    .. versionadded:: 1.1.0

    Attributes
    ----------
    world: :class:`str`
        The world the highscores belong to. If this is :obj:`None`, the highscores shown are for all worlds.
    category: :class:`Category`
        The selected category to displays the highscores of.
    vocation: :class:`VocationFilter`
        The selected vocation to filter out values.
    battleye_filter: :class:`BattlEyeHighscoresFilter`
        The selected BattlEye filter. If :obj:`None`, all worlds will be displayed.

        Only applies for global highscores. Only characters from worlds with the matching BattlEye protection will be
        shown.
    pvp_types_filter: :class:`list` of :class:`PvpTypeFilter`
        The selected PvP types filter. If :obj:`None`, all world will be displayed.

        Only applies for global highscores. Only characters from worlds with the matching PvP type will be shown.
    page: :class:`int`
        The page number being displayed.
    total_pages: :class:`int`
        The total number of pages.
    results_count: :class:`int`
        The total amount of highscores entries in this category. These may be shown in another page.
    last_updated: :class:`datetime.timedelta`
        How long ago were this results updated. The resolution is 1 minute.
    entries: :class:`list` of :class:`HighscoresEntry`
        The highscores entries found.
    available_worlds: :class:`list` of :class:`str`
        The worlds available for selection.
    """

    _ENTRIES_PER_PAGE = 50

    def __init__(self, world, category=Category.EXPERIENCE, **kwargs):
        self.world: Optional[str] = world
        self.category: Category = try_enum(Category, category, Category.EXPERIENCE)
        self.vocation: VocationFilter = try_enum(VocationFilter, kwargs.get("vocation"), VocationFilter.ALL)
        self.battleye_filter: Optional[BattlEyeTypeFilter] = try_enum(BattlEyeTypeFilter, kwargs.get("battleye_filter"))
        self.pvp_types_filter: List[PvpTypeFilter] = kwargs.get("pvp_types_filter", [])
        self.entries: List[HighscoresEntry] = kwargs.get("entries", [])
        self.results_count: int = kwargs.get("results_count", 0)
        self.page: int = kwargs.get("page", 1)
        self.total_pages: int = kwargs.get("total_pages", 1)

    __slots__ = (
        'world',
        'category',
        'vocation',
        'battleye_filter',
        "pvp_types_filter",
        'page',
        'total_pages',
        'results_count',
        'last_updated',
        'entries',
        'available_worlds',
    )

    _serializable_properties = (

    )

    def __repr__(self):
        return f"<{self.__class__.__name__} world={self.world!r} category={self.category!r} vocation={self.vocation!r}>"

    @property
    def from_rank(self):
        """:class:`int`: The starting rank of the provided entries."""
        return self.entries[0].rank if self.entries else 0

    @property
    def to_rank(self):
        """:class:`int`: The last rank of the provided entries."""
        return self.entries[-1].rank if self.entries else 0

    @property
    def url(self):
        """:class:`str`: The URL to the highscores page on Tibia.com containing the results."""
        return self.get_url(self.world, self.category, self.vocation, self.page, self.battleye_filter,
                            self.pvp_types_filter)

    @property
    def previous_page_url(self):
        """:class:`str`: The URL to the previous page of the current highscores, if there's any."""
        return self.get_page_url(self.page - 1) if self.page > 1 else None

    @property
    def next_page_url(self):
        """:class:`str`: The URL to the next page of the current highscores, if there's any."""
        return self.get_page_url(self.page + 1) if self.page < self.total_pages else None

    def get_page_url(self, page):
        """Get the URL to a specific page for the current highscores.

        Parameters
        ----------
        page: :class:`int`
            The page to get the URL for.

        Returns
        -------
        :class:`str`
            The URL to the page of the current highscores.

        Raises
        ------
        ValueError
            The provided page is less or equals than zero.
        """
        if page <= 0:
            raise ValueError("page cannot be less or equals than zero")
        return self.get_url(self.world, self.category, self.vocation, page, self.battleye_filter, self.pvp_types_filter)

    @classmethod
    def from_content(cls, content):
        """Create an instance of the class from the html content of a highscores page.

        Notes
        -----
        Tibia.com only shows up to 50 entries per page, so in order to obtain the full highscores, all pages must be
        obtained individually and merged into one.

        Parameters
        ----------
        content: :class:`str`
            The HTML content of the page.

        Returns
        -------
        :class:`Highscores`
            The highscores results contained in the page.

        Raises
        ------
        InvalidContent
            If content is not the HTML of a highscore's page.
        """
        parsed_content = parse_tibiacom_content(content)
        form = parsed_content.find("form")
        tables = cls._parse_tables(parsed_content)
        if form is None:
            if "Error" in tables and "The world doesn't exist!" in tables["Error"].text:
                return None
            raise InvalidContent("content does is not from the highscores section of Tibia.com")
        highscores = cls(None)
        highscores._parse_filters_table(form)
        last_update_container = parsed_content.find("span", attrs={"class": "RightArea"})
        if last_update_container:
            m = numeric_pattern.search(last_update_container.text)
            highscores.last_updated = datetime.timedelta(minutes=int(m.group(1))) if m else datetime.timedelta()
        entries_table = tables.get("Highscores")
        highscores._parse_entries_table(entries_table)
        return highscores

    @classmethod
    def get_url(cls, world=None, category=Category.EXPERIENCE, vocation=VocationFilter.ALL, page=1,
                battleye_type=None, pvp_types=None):
        """Get the Tibia.com URL of the highscores for the given parameters.

        Parameters
        ----------
        world: :class:`str`, optional
            The game world of the desired highscores. If no world is passed, ALL worlds are shown.
        category: :class:`Category`
            The desired highscores category.
        vocation: :class:`VocationFilter`
            The vocation filter to apply. By default all vocations will be shown.
        page: :class:`int`
            The page of highscores to show.
        battleye_type: :class:`BattlEyeHighscoresFilter`, optional
            The battleEye filters to use.
        pvp_types: :class:`list` of :class:`PvpTypeFilter`, optional
            The list of PvP types to filter the results for.

        Returns
        -------
        The URL to the Tibia.com highscores.
        """
        pvp_types = pvp_types or []
        pvp_params = [("worldtypes[]", p.value) for p in pvp_types]
        return get_tibia_url("community", "highscores", *pvp_params, world=world, category=category.value,
                             profession=vocation.value, currentpage=page,
                             beprotection=battleye_type.value if battleye_type else None)

    # region Private methods
    def _parse_entries_table(self, table):
        """Parse the table containing the highscore entries.

        Parameters
        ----------
        table: :class:`bs4.Tag`
            The table containing the entries.
        """
        entries = table.find_all("tr")
        if entries is None:
            return
        _, header, *rows = entries
        info_row = rows.pop()
        pages_div, results_div = info_row.find_all("div")
        page_links = pages_div.find_all("a")
        listed_pages = [int(p.text) for p in page_links]
        if listed_pages:
            self.page = next((x for x in range(1, listed_pages[-1] + 1) if x not in listed_pages), listed_pages[-1] + 1)
            self.total_pages = max(int(page_links[-1].text), self.page)
        self.results_count = parse_integer(results_pattern.search(results_div.text).group(1))
        for row in rows:
            cols_raw = row.find_all('td')
            if "There is currently no data" in cols_raw[0].text:
                break
            if len(cols_raw) <= 2:
                break
            self._parse_entry(cols_raw)

    def _parse_filters_table(self, form):
        """
        Parse the filters table found in a highscores page.

        Parameters
        ----------
        form: :class:`bs4.Tag`
            The table containing the filters.
        """
        data = parse_form_data(form, include_options=True)
        self.world = data["world"] if data.get("world") else None
        self.battleye_filter = try_enum(BattlEyeHighscoresFilter, parse_integer(data.get("beprotection"), None))
        self.category = try_enum(Category, parse_integer(data.get("category"), None))
        self.vocation = try_enum(VocationFilter, parse_integer(data.get("profession"), None), VocationFilter.ALL)
        checkboxes = form.find_all("input", {"type": "checkbox", "checked": "checked"})
        values = [int(c["value"]) for c in checkboxes]
        self.pvp_types_filter = [try_enum(PvpTypeFilter, v) for v in values]
        self.available_words = [v for v in data["__options__"]["world"].values() if v]

    @classmethod
    def _parse_tables(cls, parsed_content):
        """
        Parse the information tables found in a highscores page.

        Parameters
        ----------
        parsed_content: :class:`bs4.BeautifulSoup`
            A :class:`BeautifulSoup` object containing all the content.

        Returns
        -------
        :class:`OrderedDict`[:class:`str`, :class:`bs4.Tag`]
            A dictionary containing all the table rows, with the table headers as keys.
        """
        tables = parsed_content.find_all('div', attrs={'class': 'TableContainer'})
        output = OrderedDict()
        for table in tables:
            title = table.find("div", attrs={'class': 'Text'}).text
            title = title.split("[")[0].strip()
            title = re.sub(r'Last Update.*', '', title)
            inner_table = table.find("div", attrs={'class': 'InnerTableContainer'})
            output[title] = inner_table
        return output

    def _parse_entry(self, cols):
        """Parse an entry's row and adds the result to py:attr:`entries`.

        Parameters
        ----------
        cols: :class:`bs4.ResultSet`
            The list of columns for that entry.
        """
        rank, name, *values = [c.text.replace('\xa0', ' ').strip() for c in cols]
        rank = int(rank)
        extra = None
        if self.category == Category.LOYALTY_POINTS:
            extra, vocation, world, level, value = values
        else:
            vocation, world, level, value = values
        value = int(value.replace(',', ''))
        level = int(level)
        if self.category == Category.LOYALTY_POINTS:
            entry = LoyaltyHighscoresEntry(rank, name, vocation, world, level, value, extra)
        else:
            entry = HighscoresEntry(rank, name, vocation, world, level, value)
        self.entries.append(entry)
    # endregion


class HighscoresEntry(abc.BaseCharacter, abc.Serializable):
    """Represents a entry for the highscores.

    Attributes
    ----------
    name: :class:`str`
        The name of the character.
    rank: :class:`int`
        The character's rank in the respective highscores.
    vocation: :class:`Vocation`
        The character's vocation.
    world: :class:`str`
        The character's world.
    level: :class:`int`
        The character's level.
    value: :class:`int`
        The character's value for the highscores.
    """

    def __init__(self, rank, name, vocation, world, level, value):
        self.name: str = name
        self.rank: int = rank
        self.vocation = try_enum(Vocation, vocation)
        self.value: int = value
        self.world: str = world
        self.level: int = level

    __slots__ = (
        'rank',
        'name',
        'vocation',
        'world',
        'level',
        'value',
    )

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} rank={self.rank} name={self.name!r} value={self.value}>"


class LoyaltyHighscoresEntry(HighscoresEntry):
    """Represents a entry for the highscores loyalty points category.

    This is a subclass of :class:`HighscoresEntry`, adding an extra field for title.

    Attributes
    ----------
    name: :class:`str`
        The name of the character.
    rank: :class:`int`
        The character's rank in the respective highscores.
    vocation: :class:`Vocation`
        The character's vocation.
    world: :class:`str`
        The character's world.
    level: :class:`int`
        The character's level.
    value: :class:`int`
        The character's loyalty points.
    title: :class:`str`
        The character's loyalty title.
    """

    def __init__(self, rank, name, vocation, world, level, value, title):
        super().__init__(rank, name, vocation, world, level, value)
        self.title: str = title

    __slots__ = (
        'title',
    )



"""Exceptions thrown by tibia.py."""


class TibiapyException(Exception):
    """Base exception for the tibiapy module.

    All exceptions thrown by the module are inherited from this.
    """

    pass


class InvalidContent(TibiapyException):
    """Exception thrown when the provided content is unrelated for the calling function.

    This usually means that the content provided belongs to a different website or section of the website.
    This serves as a way to differentiate those cases from a parsing that returned no results (e.g. Character not found)

    In some cases this can mean that Tibia.com's format has changed and the library needs updating.

    Attributes
    ----------
    original: :class:`Exception`
        The original exception that caused this exception.
    """

    def __init__(self, message, original=None):
        super().__init__(message)
        self.original = original


class NetworkError(TibiapyException):
    """Exception thrown when there was a network error trying to fetch a resource from the web.

    Attributes
    ----------
    original: :class:`Exception`
        The original exception that caused this exception.
    fetching_time: :class:`float`
        The time between the request and the response.
    """

    def __init__(self, message, original=None, fetching_time=0):
        super().__init__(message)
        self.original = original
        self.fetching_time = fetching_time


class Forbidden(NetworkError):
    """A subclass of :class:`NetworkError` thrown when Tibia.com returns a 403 status code.

    Tibia.com returns a 403 status code when it detects that too many requests are being done.
    This has its own subclass to let the user decide to treat this differently than other network errors.
    """


class SiteMaintenanceError(NetworkError):
    """A subclass of :class:`NetworkError` thrown when Tibia.com is down for maintenance.

    When Tibia.com is under maintenance, all sections of the website redirect to maintenance.tibia.com.
    """



import datetime
import unittest

from tests.tests_tibiapy import TestCommons
from tibiapy import Category, Highscores, HighscoresEntry, InvalidContent, LoyaltyHighscoresEntry, \
    Vocation, VocationFilter, BattlEyeHighscoresFilter

FILE_HIGHSCORES_FULL = "highscores/tibiacom_full.txt"
FILE_HIGHSCORES_EXPERIENCE = "highscores/tibiacom_experience.txt"
FILE_HIGHSCORES_LOYALTY = "highscores/tibiacom_loyalty.txt"
FILE_HIGHSCORES_BATTLEYE_PVP_FILTER = "highscores/tibiacom_battleye_pvp_filters.txt"
FILE_HIGHSCORES_EMPTY = "highscores/tibiacom_empty.txt"
FILE_HIGHSCORES_NO_RESULTS = "highscores/tibiacom_no_results.txt"


class TestHighscores(unittest.TestCase, TestCommons):
    # region Tibia.com Tests
    def test_highscores_from_content(self):
        """Testing parsing Highscores"""
        content = self.load_resource(FILE_HIGHSCORES_FULL)
        highscores = Highscores.from_content(content)

        self.assertEqual("Estela", highscores.world)
        self.assertEqual(VocationFilter.KNIGHTS, highscores.vocation)
        self.assertEqual(Category.MAGIC_LEVEL, highscores.category)
        self.assertEqual(BattlEyeHighscoresFilter.ANY_WORLD, highscores.battleye_filter)
        self.assertEqual(1983, highscores.results_count)
        self.assertEqual(38, highscores.from_rank)
        self.assertEqual(38, highscores.to_rank)
        self.assertEqual(4, highscores.page)
        self.assertEqual(40, highscores.total_pages)
        self.assertIsNotNone(highscores.get_page_url(1))
        self.assertIsNotNone(highscores.url)
        self.assertIsNotNone(highscores.next_page_url)
        self.assertIsNotNone(highscores.previous_page_url)
        self.assertEqual(datetime.timedelta(minutes=6), highscores.last_updated)

        for entry in highscores.entries:
            self.assertIsInstance(entry, HighscoresEntry)
            self.assertIsInstance(entry.name, str)
            self.assertIsInstance(entry.vocation, Vocation)
            self.assertIsInstance(entry.rank, int)
            self.assertIsInstance(entry.value, int)
            self.assertIsInstance(entry.level, int)
            self.assertEqual("Estela", entry.world)

    def test_highscores_from_content_highscores(self):
        """Testing parsing experience highscores"""
        content = self.load_resource(FILE_HIGHSCORES_EXPERIENCE)
        highscores = Highscores.from_content(content)

        self.assertEqual(highscores.world, "Gladera")
        self.assertEqual(highscores.vocation, VocationFilter.PALADINS)
        self.assertEqual(highscores.category, Category.EXPERIENCE)
        self.assertEqual(highscores.results_count, 1000)
        self.assertEqual(highscores.total_pages, 20)

        for entry in highscores.entries:
            self.assertIsInstance(entry, HighscoresEntry)
            self.assertIsInstance(entry.name, str)
            self.assertIsInstance(entry.vocation, Vocation)
            self.assertIsInstance(entry.rank, int)
            self.assertIsInstance(entry.value, int)
            self.assertIsInstance(entry.level, int)
            self.assertEqual(entry.world, "Gladera")

    def test_highscores_from_content_loyalty(self):
        """Testing parsing experience loyalty"""
        content = self.load_resource(FILE_HIGHSCORES_LOYALTY)
        highscores = Highscores.from_content(content)

        self.assertEqual("Calmera", highscores.world)
        self.assertEqual(VocationFilter.PALADINS, highscores.vocation)
        self.assertEqual(Category.LOYALTY_POINTS, highscores.category)
        self.assertEqual(1007, highscores.results_count)
        self.assertEqual(21, highscores.total_pages)

        for entry in highscores.entries:
            self.assertIsInstance(entry, LoyaltyHighscoresEntry)
            self.assertIsInstance(entry.name, str)
            self.assertIsInstance(entry.vocation, Vocation)
            self.assertIsInstance(entry.rank, int)
            self.assertIsInstance(entry.value, int)
            self.assertIsInstance(entry.level, int)
            self.assertIsInstance(entry.title, str)

    def test_highscores_from_content_battleye_and_pvp_filters(self):
        """Testing parsing Highscores"""
        content = self.load_resource(FILE_HIGHSCORES_BATTLEYE_PVP_FILTER)
        highscores = Highscores.from_content(content)

        self.assertEqual(None, highscores.world)
        self.assertEqual(VocationFilter.ALL, highscores.vocation)
        self.assertEqual(Category.EXPERIENCE, highscores.category)
        self.assertEqual(BattlEyeHighscoresFilter.INITIALLY_PROTECTED, highscores.battleye_filter)
        self.assertEqual(1000, highscores.results_count)
        self.assertEqual(1, highscores.from_rank)
        self.assertEqual(50, highscores.to_rank)
        self.assertEqual(1, highscores.page)
        self.assertEqual(20, highscores.total_pages)
        self.assertEqual(3, len(highscores.pvp_types_filter))
        self.assertIsNotNone(highscores.url)
        self.assertEqual(datetime.timedelta(minutes=27), highscores.last_updated)

        for entry in highscores.entries:
            self.assertIsInstance(entry, HighscoresEntry)
            self.assertIsInstance(entry.name, str)
            self.assertIsInstance(entry.vocation, Vocation)
            self.assertIsInstance(entry.rank, int)
            self.assertIsInstance(entry.value, int)
            self.assertIsInstance(entry.level, int)

    def test_highscores_from_content_empty(self):
        """Testing parsing highscores when empty (world doesn't exist)"""
        content = self.load_resource(FILE_HIGHSCORES_EMPTY)
        highscores = Highscores.from_content(content)

        self.assertIsNone(highscores)

    def _test_highscores_from_content_no_results(self):
        """Testing parsing highscores with no results (first day of a new world)"""
        content = self.load_resource(FILE_HIGHSCORES_NO_RESULTS)
        highscores = Highscores.from_content(content)

        self.assertIsInstance(highscores, Highscores)
        self.assertEqual(highscores.world, "Unica")
        self.assertEqual(highscores.category, Category.EXPERIENCE)
        self.assertEqual(highscores.vocation, VocationFilter.ALL)
        self.assertEqual(highscores.total_pages, 0)
        self.assertEqual(len(highscores.entries), 0)

    def test_highscores_from_content_unrelated_section(self):
        """Testing parsing an unrelated section"""
        content = self.load_resource(self.FILE_UNRELATED_SECTION)
        with self.assertRaises(InvalidContent):
            Highscores.from_content(content)

    # endregion
